#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging

from libnmstate.error import NmstateLibnmError
from libnmstate.schema import InterfaceType

from .common import NM
from .macvlan import is_macvtap
from .translator import Nm2Api
from .veth import is_veth


NM_DBUS_INTERFACE_DEVICE = f"{NM.DBUS_INTERFACE}.Device"
NM_USE_DEFAULT_TIMEOUT_VALUE = -1


class DeviceReapply:
    def __init__(
        self,
        ctx,
        iface_name,
        iface_type,
        nm_dev,
        nm_simple_conn,
        profile_activation,
    ):
        self._ctx = ctx
        self._iface_name = iface_name
        self._iface_type = iface_type
        self._nm_dev = nm_dev
        self._nm_simple_conn = nm_simple_conn
        self._profile_activation = profile_activation

    def run(self):
        """
        Modify the given connection profile on the device without bring the
        interface down.
        If failed, fall back to normal profile activation
        """
        version_id = 0
        flags = 0
        action = (
            f"Reapply device config: {self._iface_name} {self._iface_type} "
            f"{self._nm_simple_conn.get_uuid()}"
        )
        self._ctx.register_async(action)
        user_data = action
        self._nm_dev.reapply_async(
            self._nm_simple_conn,
            version_id,
            flags,
            self._ctx.cancellable,
            self._reapply_callback,
            user_data,
        )

    def _reapply_callback(self, nm_dev, result, user_data):
        action = user_data
        if self._ctx.is_cancelled():
            return
        try:
            success = nm_dev.reapply_finish(result)
        except Exception as e:
            logging.debug(
                f"Device reapply failed on {self._iface_name} "
                f"{self._iface_type}: error={e}, "
                "Fallback to device activation"
            )
            self._ctx.finish_async(action, suppress_log=True)
            self._profile_activation.run()
            return

        if success:
            self._ctx.finish_async(action)
        else:
            logging.debug(
                "Device reapply failed, fallback to device activation: "
                f"iface={self._iface_name}, type={self._iface_type} "
                "error='None returned from reapply_finish()'"
            )
            self._ctx.finish_async(action, suppress_log=True)
            self._profile_activation.run()


class DeviceDelete:
    def __init__(self, ctx, iface_name, iface_type, nm_dev):
        self._ctx = ctx
        self._iface_name = iface_name
        self._iface_type = iface_type
        self._nm_dev = nm_dev

    def run(self):
        action = f"Delete device: {self._iface_type} {self._iface_name}"
        user_data = action
        self._ctx.register_async(action)
        self._nm_dev.delete_async(
            self._ctx.cancellable, self._delete_device_callback, user_data
        )

    def _delete_device_callback(self, nm_dev, result, user_data):
        action = user_data
        if self._ctx.is_cancelled():
            return
        error = None
        try:
            nm_dev.delete_finish(result)
        except Exception as e:
            error = e

        if not nm_dev.is_real():
            logging.debug(
                f"Interface is deleted and not real/exist anymore: "
                f"iface={self._iface_name} type={self._iface_type}"
            )
            if error:
                logging.debug(f"Ignored error: {error}")
            self._ctx.finish_async(action)
        else:
            self._ctx.fail(
                NmstateLibnmError(
                    f"{action} failed: error={error or 'unknown'}"
                )
            )


def list_devices(client):
    return client.get_devices()


def get_device_common_info(dev):
    return {
        "name": dev.get_iface(),
        "type_id": dev.get_device_type(),
        "type_name": dev.get_type_description(),
        "state": dev.get_state(),
    }


def is_externally_managed(nm_dev):
    nm_ac = nm_dev.get_active_connection()
    return nm_ac and NM.ActivationStateFlags.EXTERNAL & nm_ac.get_state_flags()


def get_iface_type(nm_dev):
    # TODO: Below code are mimic from translator, need redesign on this
    iface_type = nm_dev.get_type_description()
    if iface_type != InterfaceType.ETHERNET:
        iface_type = Nm2Api.get_iface_type(nm_dev.get_type_description())
    if iface_type == InterfaceType.MAC_VLAN:
        # Check whether we are MAC VTAP as NM is treating both of them as
        # MAC VLAN.
        # BUG: We should use applied config here.
        nm_ac = nm_dev.get_active_connection()
        if nm_ac:
            nm_profile = nm_ac.get_connection()
            if nm_profile and is_macvtap(nm_profile):
                iface_type = InterfaceType.MAC_VTAP
    elif iface_type == InterfaceType.ETHERNET:
        # Check whether we are Veth as NM is treating both of them as
        # Ethernet.
        nm_ac = nm_dev.get_active_connection()
        if nm_ac:
            nm_profile = nm_ac.get_connection()
            if nm_profile and is_veth(nm_profile):
                iface_type = InterfaceType.VETH
    return iface_type


def get_nm_dev(ctx, iface_name, iface_type):
    """
    Return the first NM.Device matching iface_name and iface_type.
    We don't use `NM.Client.get_device_by_iface()` as nm_dev does not
    kernel interface, it could be OVS bridge or OVS port where name
    can duplicate with kernel interface name.
    """
    for nm_dev in ctx.client.get_devices():
        cur_iface_type = get_iface_type(nm_dev)
        if nm_dev.get_iface() == iface_name and (
            iface_type is None or cur_iface_type == iface_type
        ):
            return nm_dev
    return None



#
# Copyright (c) 2020-2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateValueError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import MacVlan
from .common import NM


NMSTATE_MODE_TO_NM_MODE = {
    MacVlan.Mode.VEPA: NM.SettingMacvlanMode.VEPA,
    MacVlan.Mode.BRIDGE: NM.SettingMacvlanMode.BRIDGE,
    MacVlan.Mode.PRIVATE: NM.SettingMacvlanMode.PRIVATE,
    MacVlan.Mode.PASSTHRU: NM.SettingMacvlanMode.PASSTHRU,
    MacVlan.Mode.SOURCE: NM.SettingMacvlanMode.SOURCE,
}


def create_setting(iface, base_con_profile, tap=False):
    macvlan_setting = None
    if base_con_profile:
        macvlan_setting = base_con_profile.get_setting_by_name(
            NM.SETTING_MACVLAN_SETTING_NAME
        )
        if macvlan_setting:
            macvlan_setting = macvlan_setting.duplicate()

    if not macvlan_setting:
        macvlan_setting = NM.SettingMacvlan.new()

    nm_mode = NMSTATE_MODE_TO_NM_MODE.get(iface.mode)
    if not nm_mode or nm_mode == MacVlan.Mode.SOURCE:
        raise NmstateValueError(
            f"{iface.mode} is not valid or supported by " "NetworkManager"
        )

    macvlan_setting.props.mode = nm_mode
    macvlan_setting.props.parent = iface.base_iface
    macvlan_setting.props.tap = tap
    if iface.promiscuous is not None:
        macvlan_setting.props.promiscuous = iface.promiscuous

    return macvlan_setting


def is_macvtap(applied_config):
    if applied_config:
        macvlan_setting = applied_config.get_setting_by_name(
            NM.SETTING_MACVLAN_SETTING_NAME
        )
        if macvlan_setting:
            return macvlan_setting.props.tap
    return False


def get_current_macvlan_type(applied_config):
    """
    This is a workaround needed due to Nmstate gathering the interface type
    from NetworkManager, as we are deciding the interface type using the
    setting name. If the interface type is not adjusted, Nmstate will fail
    during verification as NM and Nispor interfaces will not be merged
    correctly.
    """
    if is_macvtap(applied_config):
        return {Interface.TYPE: InterfaceType.MAC_VTAP}
    return {}



#
# Copyright (c) 2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import InterfaceType
from libnmstate.schema import Veth

from .ethernet import EthernetIface


class VethIface(EthernetIface):
    def __init__(self, info, save_to_disk=True):
        super().__init__(info, save_to_disk)
        self._peer_changed = False

    @property
    def is_virtual(self):
        return True

    @property
    def peer(self):
        return self.raw.get(Veth.CONFIG_SUBTREE, {}).get(Veth.PEER)

    @property
    def is_peer_changed(self):
        return self._peer_changed

    def mark_as_peer(self):
        self._is_peer = True

    def _mark_iface_is_peer(self, ifaces):
        if (
            not self.is_absent
            and not ifaces.get_cur_iface(self.peer, self.type)
            and not ifaces.get_cur_iface(self.name, self.type)
        ):
            for iface in ifaces.all_ifaces():
                if iface.name == self.peer and (
                    self.type == iface.type
                    or iface.type == InterfaceType.ETHERNET
                ):
                    if not self.is_peer:
                        iface._is_peer = True

    def _mark_peer_changed(self, ifaces):
        if self.is_up:
            cur_iface = ifaces.get_cur_iface(self.name, self.type)
            if cur_iface and self.peer != cur_iface.peer:
                self._peer_changed = True

    def gen_metadata(self, ifaces):
        self._mark_iface_is_peer(ifaces)
        self._mark_peer_changed(ifaces)
        super().gen_metadata(ifaces)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import Route


IPV4_DEFAULT_GATEWAY_DESTINATION = "0.0.0.0/0"
IPV6_DEFAULT_GATEWAY_DESTINATION = "::/0"

LOCAL_ROUTE_TABLE = 255


def nispor_route_state_to_nmstate(np_routes):
    return [
        _nispor_route_to_nmstate(rt)
        for rt in np_routes
        if rt.scope in ["universe", "link"]
        and rt.oif != "lo"
        and rt.table != LOCAL_ROUTE_TABLE
    ]


def nispor_route_state_to_nmstate_static(np_routes):
    return [
        _nispor_route_to_nmstate(rt)
        for rt in np_routes
        if rt.scope in ["universe", "link"]
        and rt.protocol not in ["kernel", "ra", "dhcp"]
        and rt.oif != "lo"
        and rt.table != LOCAL_ROUTE_TABLE
    ]


def _nispor_route_to_nmstate(np_rt):
    if np_rt.dst:
        destination = np_rt.dst
    elif np_rt.gateway:
        destination = (
            IPV6_DEFAULT_GATEWAY_DESTINATION
            if np_rt.address_family == "ipv6"
            else IPV4_DEFAULT_GATEWAY_DESTINATION
        )
    else:
        destination = ""

    if np_rt.via:
        next_hop = np_rt.via
    elif np_rt.gateway:
        next_hop = np_rt.gateway
    else:
        next_hop = ""

    return {
        Route.TABLE_ID: np_rt.table,
        Route.DESTINATION: destination,
        Route.NEXT_HOP_INTERFACE: np_rt.oif if np_rt.oif else "",
        Route.NEXT_HOP_ADDRESS: next_hop,
        Route.METRIC: np_rt.metric if np_rt.metric else 0,
    }



#
# Copyright (c) 2018-2019 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import ipaddress
from libnmstate.error import NmstateValueError

_IPV6_LINK_LOCAL_NETWORK_PREFIXES = ["fe8", "fe9", "fea", "feb"]
_IPV6_LINK_LOCAL_NETWORK_PREFIX_LENGTH = 10

KERNEL_MAIN_ROUTE_TABLE_ID = 254


def is_ipv6_link_local_addr(ip, prefix):
    return (
        ip[: len(_IPV6_LINK_LOCAL_NETWORK_PREFIXES[0])]
        in _IPV6_LINK_LOCAL_NETWORK_PREFIXES
        and prefix >= _IPV6_LINK_LOCAL_NETWORK_PREFIX_LENGTH
    )


def is_ipv6_address(addr):
    return ":" in addr


def to_ip_address_full(ip, prefix=None):
    if prefix:
        return f"{ip}/{prefix}"
    else:
        return to_ip_address_full(*ip_address_full_to_tuple(ip))


def ip_address_full_to_tuple(addr):
    try:
        net = ipaddress.ip_network(addr)
    except (ipaddress.AddressValueError, ipaddress.NetmaskValueError) as err:
        raise NmstateValueError(f"Invalid IP address, error: {err}")

    return f"{net.network_address}", net.prefixlen


def canonicalize_ip_network(address):
    try:
        return ipaddress.ip_network(address, strict=False).with_prefixlen
    except ValueError as e:
        raise NmstateValueError(f"Invalid IP network address: {e}")


def canonicalize_ip_address(address):
    try:
        return ipaddress.ip_address(address).compressed
    except ValueError as e:
        raise NmstateValueError(f"Invalid IP address: {e}")



#
# Copyright (c) 2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateNotSupportedError
from libnmstate.ifaces.veth import VethIface
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceState
from libnmstate.schema import InterfaceType
from libnmstate.schema import Veth
from .common import NM


SETTING_VETH = "SettingVeth"
SETTING_VETH_SETTING_NAME = "SETTING_VETH_SETTING_NAME"


def create_setting(iface, base_con_profile):
    veth_setting = None

    if not is_nm_veth_supported():
        raise NmstateNotSupportedError(
            "Configuring veth interfaces is only supported on NetworkManager "
            "1.30 or greater."
        )

    if base_con_profile:
        veth_setting = base_con_profile.get_setting_by_name(
            getattr(NM, SETTING_VETH_SETTING_NAME)
        )
        if veth_setting:
            veth_setting = veth_setting.duplicate()

    if not veth_setting:
        veth_setting = getattr(NM, SETTING_VETH).new()
    veth_setting.props.peer = iface.peer

    return veth_setting


def is_nm_veth_supported():
    return hasattr(NM, SETTING_VETH_SETTING_NAME)


def is_veth(applied_config):
    if applied_config and is_nm_veth_supported():
        veth_setting = applied_config.get_setting_by_name(
            getattr(NM, SETTING_VETH_SETTING_NAME)
        )
        if veth_setting:
            return True

    return False


def get_current_veth_type(applied_config):
    """
    This is a workaround needed due to Nmstate gathering interface type from
    NetworkManager, as NetworkManager is exposing veth interfaces as ethernet.
    If the interface type is not adjusted, Nmstate will fail during
    verification as NM and Nispor interfaces will not be merged correctly.
    """
    if is_veth(applied_config):
        return {Interface.TYPE: InterfaceType.VETH}
    return {}


def create_iface_for_nm_veth_peer(iface):
    return VethIface(
        {
            Interface.NAME: iface.peer,
            Interface.STATE: InterfaceState.UP,
            Interface.TYPE: InterfaceType.VETH,
            Veth.CONFIG_SUBTREE: {Veth.PEER: iface.name},
        }
    )



#
# Copyright (c) 2018-2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

# Handle the NM.SimpleConnection related stuff

import uuid

from libnmstate.error import NmstatePluginError
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import LinuxBridge as LB
from libnmstate.schema import OVSBridge as OvsB
from libnmstate.schema import OVSInterface
from libnmstate.schema import OvsDB

from libnmstate.ifaces.bridge import BridgeIface

from .bond import create_setting as create_bond_setting
from .bridge import BRIDGE_TYPE as NM_LINUX_BRIDGE_TYPE
from .bridge import create_port_setting as create_linux_bridge_port_setting
from .bridge import create_setting as create_linux_bridge_setting
from .common import NM
from .ieee_802_1x import create_802_1x_setting
from .infiniband import create_setting as create_infiniband_setting
from .ipv4 import create_setting as create_ipv4_setting
from .ipv6 import create_setting as create_ipv6_setting
from .lldp import apply_lldp_setting
from .macvlan import create_setting as create_macvlan_setting
from .ovs import create_bridge_setting as create_ovs_bridge_setting
from .ovs import create_interface_setting as create_ovs_interface_setting
from .ovs import create_ovsdb_external_ids_setting
from .ovs import create_port_setting as create_ovs_port_setting
from .sriov import create_setting as create_sriov_setting
from .team import create_setting as create_team_setting
from .translator import Api2Nm
from .user import create_setting as create_user_setting
from .veth import create_setting as create_veth_setting
from .vlan import create_setting as create_vlan_setting
from .vrf import create_vrf_setting
from .vxlan import create_setting as create_vxlan_setting
from .wired import create_setting as create_wired_setting


class _ConnectionSetting:
    def __init__(self, con_setting=None):
        self._setting = con_setting

    def create(self, con_name, iface_name, iface_type, is_controller):
        con_setting = NM.SettingConnection.new()
        con_setting.props.id = con_name
        con_setting.props.interface_name = iface_name
        con_setting.props.uuid = str(uuid.uuid4())
        con_setting.props.type = iface_type
        con_setting.props.autoconnect = True
        con_setting.props.autoconnect_slaves = (
            NM.SettingConnectionAutoconnectSlaves.YES
            if is_controller
            else NM.SettingConnectionAutoconnectSlaves.DEFAULT
        )

        self._setting = con_setting

    def import_by_profile(self, profile, is_controller):
        base = profile.get_setting_connection()
        new = NM.SettingConnection.new()
        new.props.id = base.props.id
        new.props.interface_name = base.props.interface_name
        new.props.uuid = base.props.uuid
        new.props.type = base.props.type
        new.props.autoconnect = True
        new.props.autoconnect_slaves = (
            NM.SettingConnectionAutoconnectSlaves.YES
            if is_controller
            else NM.SettingConnectionAutoconnectSlaves.DEFAULT
        )

        self._setting = new

    def set_controller(self, controller, port_type):
        if controller is not None:
            self._setting.props.master = controller
            self._setting.props.slave_type = port_type

    def set_profile_name(self, con_name):
        self._setting.props.id = con_name

    @property
    def setting(self):
        return self._setting


def create_new_nm_simple_conn(iface, nm_profile):
    nm_iface_type = Api2Nm.get_iface_type(iface.type)
    iface_info = iface.to_dict()
    settings = [
        create_ipv4_setting(iface_info.get(Interface.IPV4), nm_profile),
        create_ipv6_setting(iface_info.get(Interface.IPV6), nm_profile),
    ]
    con_setting = _ConnectionSetting()
    if nm_profile and not is_multiconnect_profile(nm_profile):
        con_setting.import_by_profile(nm_profile, iface.is_controller)
        con_setting.set_profile_name(iface.name)
    else:
        con_setting.create(
            iface.name, iface.name, nm_iface_type, iface.is_controller
        )

    apply_lldp_setting(con_setting, iface_info)

    controller = iface.controller
    controller_type = iface.controller_type
    if controller_type == InterfaceType.LINUX_BRIDGE:
        controller_type = NM_LINUX_BRIDGE_TYPE
    con_setting.set_controller(controller, controller_type)
    settings.append(con_setting.setting)

    # Only apply wired/ethernet configuration based on original desire
    # state rather than the merged one.
    original_state_wired = {}
    if iface.is_desired:
        original_state_wired = iface.original_dict
    if iface.type != InterfaceType.INFINIBAND:
        # The IP over InfiniBand has its own setting for MTU and does not
        # have ethernet layer.
        wired_setting = create_wired_setting(original_state_wired, nm_profile)

        if wired_setting:
            settings.append(wired_setting)

    user_setting = create_user_setting(iface_info, nm_profile)
    if user_setting:
        settings.append(user_setting)

    if iface.type == InterfaceType.BOND:
        settings.append(create_bond_setting(iface, wired_setting, nm_profile))
    elif iface.type == InterfaceType.LINUX_BRIDGE:
        bridge_config = iface_info.get(LB.CONFIG_SUBTREE, {})
        bridge_options = bridge_config.get(LB.OPTIONS_SUBTREE)
        bridge_ports = bridge_config.get(LB.PORT_SUBTREE)
        if bridge_options or bridge_ports:
            linux_bridge_setting = create_linux_bridge_setting(
                iface_info,
                nm_profile,
                iface.original_dict,
            )
            settings.append(linux_bridge_setting)
    elif iface.type == InterfaceType.OVS_BRIDGE:
        ovs_bridge_state = iface_info.get(OvsB.CONFIG_SUBTREE, {})
        ovs_bridge_options = ovs_bridge_state.get(OvsB.OPTIONS_SUBTREE, {})
        if ovs_bridge_options:
            settings.append(create_ovs_bridge_setting(ovs_bridge_options))
    elif iface.type == InterfaceType.OVS_PORT:
        ovs_port_options = iface_info.get(OvsB.OPTIONS_SUBTREE)
        settings.append(create_ovs_port_setting(ovs_port_options))
    elif iface.type == InterfaceType.OVS_INTERFACE:
        patch_state = iface_info.get(OVSInterface.PATCH_CONFIG_SUBTREE)
        settings.extend(create_ovs_interface_setting(patch_state))
    elif iface.type == InterfaceType.INFINIBAND:
        ib_setting = create_infiniband_setting(
            iface_info,
            nm_profile,
            iface.original_dict,
        )
        if ib_setting:
            settings.append(ib_setting)

    bridge_port_options = iface_info.get(BridgeIface.BRPORT_OPTIONS_METADATA)
    if (
        bridge_port_options
        and iface.controller_type == InterfaceType.LINUX_BRIDGE
    ):
        settings.append(
            create_linux_bridge_port_setting(bridge_port_options, nm_profile)
        )

    vlan_setting = create_vlan_setting(iface_info, nm_profile)
    if vlan_setting:
        settings.append(vlan_setting)

    vxlan_setting = create_vxlan_setting(iface_info, nm_profile)
    if vxlan_setting:
        settings.append(vxlan_setting)

    sriov_setting = create_sriov_setting(iface_info, nm_profile)
    if sriov_setting:
        settings.append(sriov_setting)

    team_setting = create_team_setting(iface_info, nm_profile)
    if team_setting:
        settings.append(team_setting)

    if iface.type == InterfaceType.VRF:
        settings.append(create_vrf_setting(iface))

    if iface.type == InterfaceType.MAC_VLAN:
        settings.append(create_macvlan_setting(iface, nm_profile))

    if iface.type == InterfaceType.MAC_VTAP:
        settings.append(create_macvlan_setting(iface, nm_profile, tap=True))

    if iface.type == InterfaceType.VETH:
        veth_setting = create_veth_setting(iface, nm_profile)
        if veth_setting:
            settings.append(veth_setting)

    if (
        iface.controller_type
        in (
            InterfaceType.OVS_BRIDGE,
            InterfaceType.OVS_PORT,
        )
        or iface.type == InterfaceType.OVS_BRIDGE
    ):
        nm_setting = create_ovsdb_external_ids_setting(
            iface_info.get(OvsDB.OVS_DB_SUBTREE, {})
        )
        if nm_setting:
            settings.append(nm_setting)

    if iface.ieee_802_1x_conf:
        settings.append(create_802_1x_setting(iface.ieee_802_1x_conf))

    nm_simple_conn = NM.SimpleConnection.new()
    for setting in settings:
        nm_simple_conn.add_setting(setting)

    return nm_simple_conn


def nm_simple_conn_update_controller(nm_simple_conn, controller):
    nm_conn_setting = nm_simple_conn.get_setting_connection()
    if nm_conn_setting:
        nm_conn_setting.props.master = controller
    else:
        raise NmstatePluginError(
            f"BUG: NM.SimpleConnection {nm_simple_conn.get_id()}/"
            f"{nm_simple_conn.get_uuid()} "
            f"{nm_simple_conn.get_connection_type()} "
            "has no NM.SettingConnection"
        )


def nm_simple_conn_update_parent(nm_simple_conn, iface_type, parent):
    if iface_type == InterfaceType.VLAN:
        nm_setting = nm_simple_conn.get_setting_vlan()
    elif iface_type == InterfaceType.VXLAN:
        nm_setting = nm_simple_conn.get_setting_vxlan()
    elif iface_type in (InterfaceType.MAC_VLAN, InterfaceType.MAC_VTAP):
        nm_setting = nm_simple_conn.get_setting_by_name(
            NM.SETTING_MACVLAN_SETTING_NAME
        )
    elif iface_type == InterfaceType.INFINIBAND:
        nm_setting = nm_simple_conn.get_setting_infiniband()
    elif iface_type == InterfaceType.OVS_INTERFACE:
        # OVS Interface is using master
        return
    else:
        raise NmstatePluginError(
            f"BUG: NM.SimpleConnection {nm_simple_conn.get_id()}/"
            f"{nm_simple_conn.get_uuid()} "
            f"{nm_simple_conn.get_connection_type()} "
            f"shold not need parent"
        )
    nm_setting.props.parent = parent


def is_multiconnect_profile(nm_profile):
    nm_setting = nm_profile.get_setting_connection()
    return (
        nm_setting
        and nm_setting.get_multi_connect()
        == NM.ConnectionMultiConnect.MULTIPLE
    )



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateNotImplementedError
from libnmstate.error import NmstateValueError
from libnmstate.ifaces.bond import BondIface
from libnmstate.schema import Bond
from libnmstate.state import merge_dict
from .common import NM


BOND_TYPE = "bond"

SYSFS_EMPTY_VALUE = ""

NM_SUPPORTED_BOND_OPTIONS = NM.SettingBond.get_valid_options(
    NM.SettingBond.new()
)

SYSFS_BOND_OPTION_FOLDER_FMT = "/sys/class/net/{ifname}/bonding"

BOND_AD_ACTOR_SYSTEM_USE_BOND_MAC = "00:00:00:00:00:00"


def create_setting(iface, wired_setting, base_con_profile):
    bond_setting = NM.SettingBond.new()
    options = iface.original_dict.get(Bond.CONFIG_SUBTREE, {}).get(
        Bond.OPTIONS_SUBTREE
    )
    mode = iface.bond_mode
    if options != {}:
        if options is None:
            options = {}
        if not iface.is_bond_mode_changed:
            old_bond_options = {}
            if base_con_profile:
                old_bond_options = _get_bond_options_from_profiles(
                    base_con_profile.get_setting_bond()
                )
            merge_dict(options, old_bond_options)
    if mode:
        bond_setting.add_option("mode", mode)
    for option_name, option_value in options.items():
        if wired_setting and BondIface.is_mac_restricted_mode(mode, options):
            # When in MAC restricted mode, MAC address should be unset.
            wired_setting.props.cloned_mac_address = None
        if (
            option_name == "ad_actor_system"
            and option_value == BOND_AD_ACTOR_SYSTEM_USE_BOND_MAC
        ):
            # The all zero ad_actor_system is the kernel default value
            # And it is invalid to set as all zero
            continue
        if option_value != SYSFS_EMPTY_VALUE:
            option_value = _nm_fix_bond_options(option_name, option_value)
            success = bond_setting.add_option(option_name, option_value)
            if not success:
                raise NmstateValueError(
                    "Invalid bond option: '{}'='{}'".format(
                        option_name, option_value
                    )
                )

    return bond_setting


def _nm_fix_bond_options(option_name, option_value):
    if option_name == "all_slaves_active":
        if option_value == "delivered":
            option_value = 1
        elif option_value == "dropped":
            option_value = 0
        else:
            raise NmstateNotImplementedError(
                "Unsupported bond option: '{}'='{}'".format(
                    option_name, option_value
                )
            )
    elif option_name == "use_carrier":
        option_value = 1 if option_value else 0

    return str(option_value)


def _get_bond_options_from_profiles(bond_setting):
    ret = {}
    if bond_setting:
        for i in range(0, bond_setting.get_num_options()):
            name, value = bond_setting.get_option(i)[1:3]
            if name != "mode":
                ret[name] = value
    return ret



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from operator import attrgetter
import os

from libnmstate.schema import InterfaceType
from libnmstate.schema import LinuxBridge as LB

from .base_iface import NisporPluginBaseIface
from .bridge_port_vlan import get_port_vlan_info

OPT = LB.Options


NISPOR_USER_HZ_KEYS = [
    "forward_delay",
    "ageing_time",
    "hello_time",
    "max_age",
]


NISPOR_MULTICAST_ROUTERS_INT_MAP = {
    "disabled": 0,
    "temp_query": 1,
    "perm": 2,
    "temp": 3,
}

LSM_BRIDGE_OPTIONS_2_NISPOR = {
    OPT.MAC_AGEING_TIME: "ageing_time",
    OPT.GROUP_FORWARD_MASK: "group_fwd_mask",
    OPT.MULTICAST_SNOOPING: "multicast_snooping",
    OPT.HELLO_TIMER: "hello_timer",
    OPT.GC_TIMER: "gc_timer",
    OPT.MULTICAST_ROUTER: "multicast_router",
    OPT.GROUP_ADDR: "group_addr",
    OPT.HASH_MAX: "multicast_hash_max",
    OPT.MULTICAST_LAST_MEMBER_COUNT: "multicast_last_member_count",
    OPT.MULTICAST_LAST_MEMBER_INTERVAL: "multicast_last_member_interval",
    OPT.MULTICAST_QUERIER: "multicast_querier",
    OPT.MULTICAST_QUERIER_INTERVAL: "multicast_querier_interval",
    OPT.MULTICAST_QUERY_USE_IFADDR: "multicast_query_use_ifaddr",
    OPT.MULTICAST_QUERY_INTERVAL: "multicast_query_interval",
    OPT.MULTICAST_QUERY_RESPONSE_INTERVAL: "multicast_query_response_interval",
    OPT.MULTICAST_STARTUP_QUERY_COUNT: "multicast_startup_query_count",
    OPT.MULTICAST_STARTUP_QUERY_INTERVAL: "multicast_startup_query_interval",
}


LSM_BRIDGE_STP_OPTIONS_2_NISPOR = {
    LB.STP.ENABLED: "stp_state",
    LB.STP.PRIORITY: "priority",
    LB.STP.FORWARD_DELAY: "forward_delay",
    LB.STP.HELLO_TIME: "hello_time",
    LB.STP.MAX_AGE: "max_age",
}


class NisporPluginBridgeIface(NisporPluginBaseIface):
    def __init__(self, np_iface, np_ports):
        super().__init__(np_iface)
        self._np_ports = np_ports

    @property
    def type(self):
        return InterfaceType.LINUX_BRIDGE

    def _options_to_dict(self):
        np_options = self._np_iface.options
        info = _get_nispor_options(LSM_BRIDGE_OPTIONS_2_NISPOR, np_options)
        info[LB.STP_SUBTREE] = _get_nispor_options(
            LSM_BRIDGE_STP_OPTIONS_2_NISPOR, np_options
        )
        return info

    def _ports_to_dict(self):
        info = []
        for np_port in sorted(self._np_ports, key=attrgetter("name")):
            np_sub = np_port.subordinate_state
            port_info = {
                LB.Port.NAME: np_port.name,
                LB.Port.STP_HAIRPIN_MODE: np_sub.hairpin_mode,
                LB.Port.STP_PATH_COST: np_sub.stp_path_cost,
                LB.Port.STP_PRIORITY: np_sub.stp_priority,
            }
            if self._np_iface.options.get("vlan_filtering") and np_sub.vlans:
                port_info[LB.Port.VLAN_SUBTREE] = get_port_vlan_info(np_sub)

            info.append(port_info)
        return info

    def to_dict(self, config_only):
        info = super().to_dict(config_only)
        info[LB.CONFIG_SUBTREE] = {
            LB.OPTIONS_SUBTREE: self._options_to_dict(),
            LB.PORT_SUBTREE: self._ports_to_dict(),
        }
        return info


# Some keys nispor is showing as human readable value, while lsm need integer
# TODO: Use better approach after https://github.com/nispor/nispor/issues/11
def _nispor_value_to_lsm(np_name, np_value):
    """
    Return None if convertion failed.
    Return original value is convertion is not required
    """
    value = np_value
    if np_name == "multicast_router":
        value = NISPOR_MULTICAST_ROUTERS_INT_MAP.get(np_value)
    elif np_name == "stp_state":
        value = np_value in ("kernel_stp", "user_stp")
    elif np_name == "group_addr":
        value = value.upper()
    return value


def _get_nispor_options(np_prop_map, np_options):
    user_hz = os.sysconf("SC_CLK_TCK")
    info = {}
    for key, np_prop_name in np_prop_map.items():
        if np_prop_name in np_options:
            value = _nispor_value_to_lsm(
                np_prop_name, np_options[np_prop_name]
            )
            if value is None:
                continue
            if np_prop_name in NISPOR_USER_HZ_KEYS:
                value = int(value / user_hz)
            info[key] = value
    return info



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import LinuxBridge as LB

from .bridge_port_vlan import nmstate_port_vlan_to_nm
from .common import NM


BRIDGE_TYPE = "bridge"

OPT = LB.Options

NM_BRIDGE_OPTIONS_MAP = {
    OPT.GROUP_ADDR: "group_address",
    OPT.HASH_MAX: "multicast_hash_max",
    OPT.MULTICAST_LAST_MEMBER_COUNT: "multicast_last_member_count",
    OPT.MULTICAST_LAST_MEMBER_INTERVAL: "multicast_last_member_interval",
    OPT.MULTICAST_MEMBERSHIP_INTERVAL: "multicast_membership_interval",
    OPT.MULTICAST_QUERIER: "multicast_querier",
    OPT.MULTICAST_QUERIER_INTERVAL: "multicast_querier_interval",
    OPT.MULTICAST_QUERY_USE_IFADDR: "multicast_query_use_ifaddr",
    OPT.MULTICAST_QUERY_INTERVAL: "multicast_query_interval",
    OPT.MULTICAST_QUERY_RESPONSE_INTERVAL: "multicast_query_response_interval",
    OPT.MULTICAST_STARTUP_QUERY_COUNT: "multicast_startup_query_count",
    OPT.MULTICAST_STARTUP_QUERY_INTERVAL: "multicast_startup_query_interval",
}


def create_setting(
    bridge_state, base_con_profile, original_desired_iface_state
):
    options = original_desired_iface_state.get(LB.CONFIG_SUBTREE, {}).get(
        LB.OPTIONS_SUBTREE
    )
    bridge_setting = _get_current_bridge_setting(base_con_profile)
    if not bridge_setting:
        bridge_setting = NM.SettingBridge.new()

    if options:
        _set_bridge_properties(bridge_setting, options)

    bridge_setting.props.vlan_filtering = _is_vlan_filter_active(bridge_state)

    return bridge_setting


def _get_current_bridge_setting(base_con_profile):
    bridge_setting = None
    if base_con_profile:
        bridge_setting = base_con_profile.get_setting_bridge()
        if bridge_setting:
            bridge_setting = bridge_setting.duplicate()
    return bridge_setting


def _set_bridge_properties(bridge_setting, options):
    for key, val in options.items():
        if key == LB.Options.MAC_AGEING_TIME:
            bridge_setting.props.ageing_time = val
        elif key == LB.Options.GROUP_FORWARD_MASK:
            bridge_setting.props.group_forward_mask = val
        elif key == LB.Options.MULTICAST_SNOOPING:
            bridge_setting.props.multicast_snooping = val
        elif key == LB.STP_SUBTREE:
            _set_bridge_stp_properties(bridge_setting, val)
        elif key in NM_BRIDGE_OPTIONS_MAP:
            nm_prop_name = NM_BRIDGE_OPTIONS_MAP[key]
            # NM is using the sysfs name
            if key == LB.Options.GROUP_ADDR:
                val = val.lower()
            setattr(bridge_setting.props, nm_prop_name, val)


def _set_bridge_stp_properties(bridge_setting, bridge_stp):
    bridge_setting.props.stp = bridge_stp[LB.STP.ENABLED]
    if bridge_stp[LB.STP.ENABLED] is True:
        for stp_key, stp_val in bridge_stp.items():
            if stp_key == LB.STP.PRIORITY:
                bridge_setting.props.priority = stp_val
            elif stp_key == LB.STP.FORWARD_DELAY:
                bridge_setting.props.forward_delay = stp_val
            elif stp_key == LB.STP.HELLO_TIME:
                bridge_setting.props.hello_time = stp_val
            elif stp_key == LB.STP.MAX_AGE:
                bridge_setting.props.max_age = stp_val


def _is_vlan_filter_active(bridge_state):
    return any(
        port.get(LB.Port.VLAN_SUBTREE, {}) != {}
        for port in bridge_state.get(LB.CONFIG_SUBTREE, {}).get(
            LB.PORT_SUBTREE, []
        )
    )


def create_port_setting(options, base_con_profile):
    port_setting = None
    if base_con_profile:
        port_setting = base_con_profile.get_setting_bridge_port()
        if port_setting:
            port_setting = port_setting.duplicate()

    if not port_setting:
        port_setting = NM.SettingBridgePort.new()

    for key, val in options.items():
        if key == LB.Port.STP_PRIORITY:
            port_setting.props.priority = val
        elif key == LB.Port.STP_HAIRPIN_MODE:
            port_setting.props.hairpin_mode = val
        elif key == LB.Port.STP_PATH_COST:
            port_setting.props.path_cost = val
        elif key == LB.Port.VLAN_SUBTREE:
            port_setting.clear_vlans()
            for vlan_config in nmstate_port_vlan_to_nm(val):
                port_setting.add_vlan(vlan_config)

    return port_setting


def get_port(nm_device):
    return nm_device.get_slaves()



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import InterfaceType
from libnmstate.schema import MacVlan

from .base_iface import NisporPluginBaseIface


MACVLAN_FLAG_NOPROMISC = 1


MACVLAN_MODES = {
    "unknown": MacVlan.Mode.UNKNOWN,
    "vepa": MacVlan.Mode.VEPA,
    "bridge": MacVlan.Mode.BRIDGE,
    "private": MacVlan.Mode.PRIVATE,
    "passthru": MacVlan.Mode.PASSTHRU,
    "source": MacVlan.Mode.SOURCE,
}


class NisporPluginMacVlanIface(NisporPluginBaseIface):
    @property
    def type(self):
        return InterfaceType.MAC_VLAN

    def to_dict(self, config_only):
        info = super().to_dict(config_only)
        info[MacVlan.CONFIG_SUBTREE] = {
            MacVlan.BASE_IFACE: self._np_iface.base_iface,
            MacVlan.MODE: MACVLAN_MODES.get(
                self._np_iface.mode, MacVlan.Mode.UNKNOWN
            ),
            MacVlan.PROMISCUOUS: self._flag_to_promiscuous(
                self._np_iface.mac_vlan_flags
            ),
        }

        return info

    @staticmethod
    def _flag_to_promiscuous(flags):
        return (flags & MACVLAN_FLAG_NOPROMISC) == 0



#
# Copyright (c) 2021 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateLibnmError
from libnmstate.schema import Ieee8021X

from .common import NM
from .common import GLib

_NM_PROP_MAP = {
    NM.SETTING_802_1X_IDENTITY: Ieee8021X.IDENTITY,
    NM.SETTING_802_1X_EAP: Ieee8021X.EAP_METHODS,
    NM.SETTING_802_1X_PRIVATE_KEY: Ieee8021X.PRIVATE_KEY,
    NM.SETTING_802_1X_CLIENT_CERT: Ieee8021X.CLIENT_CERT,
    NM.SETTING_802_1X_CA_CERT: Ieee8021X.CA_CERT,
}


def get_802_1x_info(context, nm_ac):
    if not nm_ac:
        return {}
    nm_profile = nm_ac.get_connection()
    if not nm_profile:
        return {}
    nm_setting = nm_profile.get_setting_802_1x()
    if not nm_setting:
        return {}

    secrets = _get_secrets(context, nm_profile)

    info = {}
    for nm_prop, key_name in _NM_PROP_MAP.items():
        value = getattr(nm_setting.props, nm_prop)
        if isinstance(value, GLib.Bytes):
            value = _file_path_from_glib_bytes(value)
        info[key_name] = value

    if secrets.get(NM.SETTING_802_1X_PRIVATE_KEY_PASSWORD):
        info[Ieee8021X.PRIVATE_KEY_PASSWORD] = secrets[
            NM.SETTING_802_1X_PRIVATE_KEY_PASSWORD
        ]

    return {Ieee8021X.CONFIG_SUBTREE: info}


def create_802_1x_setting(ieee_802_1x_conf):
    nm_setting = NM.Setting8021x.new()

    for nm_prop, key_name in (
        (NM.SETTING_802_1X_IDENTITY, Ieee8021X.IDENTITY),
        (NM.SETTING_802_1X_EAP, Ieee8021X.EAP_METHODS),
        (
            NM.SETTING_802_1X_PRIVATE_KEY_PASSWORD,
            Ieee8021X.PRIVATE_KEY_PASSWORD,
        ),
    ):
        if key_name in ieee_802_1x_conf:
            setattr(nm_setting.props, nm_prop, ieee_802_1x_conf[key_name])

    for nm_prop, key_name in (
        (NM.SETTING_802_1X_PRIVATE_KEY, Ieee8021X.PRIVATE_KEY),
        (NM.SETTING_802_1X_CA_CERT, Ieee8021X.CA_CERT),
        (NM.SETTING_802_1X_CLIENT_CERT, Ieee8021X.CLIENT_CERT),
    ):
        if key_name in ieee_802_1x_conf:
            setattr(
                nm_setting.props,
                nm_prop,
                _file_path_to_glib_bytes(ieee_802_1x_conf[key_name]),
            )

    return nm_setting


def _file_path_from_glib_bytes(nm_bytes):
    file_path = nm_bytes.get_data().decode("utf8")
    if file_path.startswith("file://"):
        # Black is conflicting with flake for below line:
        #   https://github.com/psf/black/issues/315
        file_path = file_path[len("file://") : (-len("\\0") + 1)]  # noqa: E203

    return file_path


def _file_path_to_glib_bytes(file_path):
    file_path_bytes = bytearray(f"file://{file_path}".encode("utf8"))
    file_path_bytes.append(0)
    return GLib.Bytes.new(file_path_bytes)


def _get_secrets(context, nm_profile):
    secrets = {}
    action = f"Retrieve 802.1x secrets of profile {nm_profile.get_uuid()}"
    context.register_async(action, fast=True)
    user_data = (context, secrets, action)
    nm_profile.get_secrets_async(
        NM.SETTING_802_1X_SETTING_NAME,
        context.cancellable,
        _get_secrets_callback,
        user_data,
    )
    context.wait_all_finish()
    return secrets


def _get_secrets_callback(nm_profile, result, user_data):
    context, secrets, action = user_data

    try:
        nm_secrets = nm_profile.get_secrets_finish(result)
    except GLib.Error as e:
        context.fail(NmstateLibnmError(f"{action} failed: error={e}"))

    context.finish_async(action)
    secrets.update(nm_secrets[NM.SETTING_802_1X_SETTING_NAME])



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import warnings

import pkgutil
import yaml


def load(schema_name):
    return yaml.load(
        pkgutil.get_data("libnmstate", "schemas/" + schema_name + ".yaml"),
        Loader=yaml.SafeLoader,
    )


ifaces_schema = load("operational-state")


class Interface:
    KEY = "interfaces"

    NAME = "name"
    TYPE = "type"
    STATE = "state"
    DESCRIPTION = "description"

    IPV4 = "ipv4"
    IPV6 = "ipv6"

    MAC = "mac-address"
    MTU = "mtu"
    COPY_MAC_FROM = "copy-mac-from"


class Route:
    KEY = "routes"

    RUNNING = "running"
    CONFIG = "config"
    STATE = "state"
    STATE_ABSENT = "absent"
    TABLE_ID = "table-id"
    DESTINATION = "destination"
    NEXT_HOP_INTERFACE = "next-hop-interface"
    NEXT_HOP_ADDRESS = "next-hop-address"
    METRIC = "metric"
    USE_DEFAULT_METRIC = -1
    USE_DEFAULT_ROUTE_TABLE = 0


class RouteRule:
    KEY = "route-rules"
    CONFIG = "config"
    IP_FROM = "ip-from"
    IP_TO = "ip-to"
    PRIORITY = "priority"
    ROUTE_TABLE = "route-table"
    USE_DEFAULT_PRIORITY = -1
    USE_DEFAULT_ROUTE_TABLE = 0
    STATE = "state"
    STATE_ABSENT = "absent"


class DNS:
    KEY = "dns-resolver"
    RUNNING = "running"
    CONFIG = "config"
    SERVER = "server"
    SEARCH = "search"


class Constants:
    INTERFACES = Interface.KEY
    ROUTES = Route.KEY
    DNS = DNS.KEY


class InterfaceState:
    KEY = Interface.STATE

    DOWN = "down"
    UP = "up"
    ABSENT = "absent"
    IGNORE = "ignore"


class InterfaceType:
    KEY = Interface.TYPE

    BOND = "bond"
    DUMMY = "dummy"
    ETHERNET = "ethernet"
    LINUX_BRIDGE = "linux-bridge"
    MAC_VLAN = "mac-vlan"
    MAC_VTAP = "mac-vtap"
    OVS_BRIDGE = "ovs-bridge"
    OVS_INTERFACE = "ovs-interface"
    OVS_PORT = "ovs-port"
    UNKNOWN = "unknown"
    VLAN = "vlan"
    VXLAN = "vxlan"
    TEAM = "team"
    VRF = "vrf"
    INFINIBAND = "infiniband"
    VETH = "veth"
    OTHER = "other"

    VIRT_TYPES = (
        BOND,
        DUMMY,
        LINUX_BRIDGE,
        OVS_BRIDGE,
        OVS_PORT,
        OVS_INTERFACE,
        TEAM,
        VETH,
        VLAN,
        VXLAN,
    )


class InterfaceIP:
    ENABLED = "enabled"
    ADDRESS = "address"
    ADDRESS_IP = "ip"
    ADDRESS_PREFIX_LENGTH = "prefix-length"
    DHCP = "dhcp"
    AUTO_DNS = "auto-dns"
    AUTO_GATEWAY = "auto-gateway"
    AUTO_ROUTES = "auto-routes"
    AUTO_ROUTE_TABLE_ID = "auto-route-table-id"


class InterfaceIPv4(InterfaceIP):
    pass


class InterfaceIPv6(InterfaceIP):
    AUTOCONF = "autoconf"


OVS_BRIDGE = "OVSBridge.Port.LinkAggregation"


DEPRECATED_CONSTANTS = {
    "Bond.SLAVES": "Bond.PORT",
    "LinkAggregation.SLAVES_SUBTREE": f"{OVS_BRIDGE}.PORT_SUBTREE",
    "LinkAggregation.Slave": f"{OVS_BRIDGE}.Port",
}


class _DeprecatorType(type):
    def __getattribute__(cls, attribute):
        try:
            return super().__getattribute__(attribute)
        except AttributeError:
            deprecated_class = cls
            deprecated_classname = deprecated_class.__name__
            deprecated_name = attribute

            oldconstant = f"{deprecated_classname}.{deprecated_name}"
            newconstant = DEPRECATED_CONSTANTS.get(oldconstant)

            if newconstant:
                warnings.warn(
                    f"Using '{oldconstant}' is deprecated, "
                    f"use '{newconstant}' instead.",
                    FutureWarning,
                    stacklevel=3,
                )

                attributes = newconstant.split(".")
                new_classname = attributes.pop(0)
                new_value = globals()[new_classname]
                while attributes:
                    new_value = getattr(new_value, attributes.pop(0))

                return new_value

            raise


class Bond(metaclass=_DeprecatorType):
    KEY = InterfaceType.BOND
    CONFIG_SUBTREE = "link-aggregation"

    MODE = "mode"
    PORT = "port"
    OPTIONS_SUBTREE = "options"


class BondMode:
    ROUND_ROBIN = "balance-rr"
    ACTIVE_BACKUP = "active-backup"
    XOR = "balance-xor"
    BROADCAST = "broadcast"
    LACP = "802.3ad"
    TLB = "balance-tlb"
    ALB = "balance-alb"


class Bridge:
    CONFIG_SUBTREE = "bridge"
    OPTIONS_SUBTREE = "options"
    PORT_SUBTREE = "port"

    class Port:
        NAME = "name"
        VLAN_SUBTREE = "vlan"

        class Vlan:
            ENABLE_NATIVE = "enable-native"
            TRUNK_TAGS = "trunk-tags"
            MODE = "mode"
            TAG = "tag"

            class Mode:
                ACCESS = "access"
                TRUNK = "trunk"
                UNKNOWN = "unknown"

            class TrunkTags:
                ID = "id"
                ID_RANGE = "id-range"
                MIN_RANGE = "min"
                MAX_RANGE = "max"


class LinuxBridge(Bridge):
    TYPE = "linux-bridge"
    STP_SUBTREE = "stp"
    MULTICAST_SUBTREE = "multicast"

    class Options:
        GROUP_FORWARD_MASK = "group-forward-mask"
        MAC_AGEING_TIME = "mac-ageing-time"
        MULTICAST_SNOOPING = "multicast-snooping"
        GROUP_ADDR = "group-addr"
        GROUP_FWD_MASK = "group-fwd-mask"
        HASH_ELASTICITY = "hash-elasticity"
        HASH_MAX = "hash-max"
        MULTICAST_ROUTER = "multicast-router"
        MULTICAST_LAST_MEMBER_COUNT = "multicast-last-member-count"
        MULTICAST_LAST_MEMBER_INTERVAL = "multicast-last-member-interval"
        MULTICAST_MEMBERSHIP_INTERVAL = "multicast-membership-interval"
        MULTICAST_QUERIER = "multicast-querier"
        MULTICAST_QUERIER_INTERVAL = "multicast-querier-interval"
        MULTICAST_QUERY_USE_IFADDR = "multicast-query-use-ifaddr"
        MULTICAST_QUERY_INTERVAL = "multicast-query-interval"
        MULTICAST_QUERY_RESPONSE_INTERVAL = "multicast-query-response-interval"
        MULTICAST_STARTUP_QUERY_COUNT = "multicast-startup-query-count"
        MULTICAST_STARTUP_QUERY_INTERVAL = "multicast-startup-query-interval"

        # Read only properties begin
        HELLO_TIMER = "hello-timer"
        GC_TIMER = "gc-timer"
        # Read only properties end

    class Port(Bridge.Port):
        STP_HAIRPIN_MODE = "stp-hairpin-mode"
        STP_PATH_COST = "stp-path-cost"
        STP_PRIORITY = "stp-priority"

    class STP:
        ENABLED = "enabled"
        FORWARD_DELAY = "forward-delay"
        HELLO_TIME = "hello-time"
        MAX_AGE = "max-age"
        PRIORITY = "priority"


class Ethernet:
    TYPE = InterfaceType.ETHERNET
    CONFIG_SUBTREE = "ethernet"

    AUTO_NEGOTIATION = "auto-negotiation"
    SPEED = "speed"
    DUPLEX = "duplex"

    FULL_DUPLEX = "full"
    HALF_DUPLEX = "half"

    SRIOV_SUBTREE = "sr-iov"

    class SRIOV:
        TOTAL_VFS = "total-vfs"
        VFS_SUBTREE = "vfs"

        class VFS:
            ID = "id"
            MAC_ADDRESS = "mac-address"
            SPOOF_CHECK = "spoof-check"
            TRUST = "trust"
            MIN_TX_RATE = "min-tx-rate"
            MAX_TX_RATE = "max-tx-rate"


class Veth:
    TYPE = InterfaceType.VETH
    CONFIG_SUBTREE = "veth"

    PEER = "peer"


class VLAN:
    TYPE = InterfaceType.VLAN
    CONFIG_SUBTREE = "vlan"

    ID = "id"
    BASE_IFACE = "base-iface"


class VXLAN:
    TYPE = InterfaceType.VXLAN
    CONFIG_SUBTREE = "vxlan"

    ID = "id"
    BASE_IFACE = "base-iface"
    REMOTE = "remote"
    DESTINATION_PORT = "destination-port"


class OvsDB:
    KEY = "ovs-db"
    OVS_DB_SUBTREE = "ovs-db"
    # Don't use hypen as this is OVS data base entry
    EXTERNAL_IDS = "external_ids"
    OTHER_CONFIG = "other_config"


class OVSInterface(OvsDB):
    TYPE = InterfaceType.OVS_INTERFACE
    PATCH_CONFIG_SUBTREE = "patch"

    class Patch:
        PEER = "peer"


class OVSBridge(Bridge, OvsDB):
    TYPE = "ovs-bridge"

    class Options:
        FAIL_MODE = "fail-mode"
        MCAST_SNOOPING_ENABLED = "mcast-snooping-enable"
        RSTP = "rstp"
        STP = "stp"

    class Port(Bridge.Port):
        LINK_AGGREGATION_SUBTREE = "link-aggregation"

        class LinkAggregation(metaclass=_DeprecatorType):
            MODE = "mode"
            PORT_SUBTREE = "port"

            class Port:
                NAME = "name"

            class Options:
                DOWN_DELAY = "bond-downdelay"
                UP_DELAY = "bond-updelay"

            class Mode:
                ACTIVE_BACKUP = "active-backup"
                BALANCE_SLB = "balance-slb"
                BALANCE_TCP = "balance-tcp"
                LACP = "lacp"


class Team:
    TYPE = InterfaceType.TEAM
    CONFIG_SUBTREE = InterfaceType.TEAM

    PORT_SUBTREE = "port"
    RUNNER_SUBTREE = "runner"

    class Port:
        NAME = "name"

    class Runner:
        NAME = "name"

        class RunnerMode:
            LOAD_BALANCE = "loadbalance"


class LLDP:
    CONFIG_SUBTREE = "lldp"
    ENABLED = "enabled"
    NEIGHBORS_SUBTREE = "neighbors"

    class Neighbors:
        DESCRIPTION = "_description"
        TLV_TYPE = "type"
        TLV_SUBTYPE = "subtype"
        ORGANIZATION_CODE = "oui"


class VRF:
    CONFIG_SUBTREE = "vrf"
    PORT_SUBTREE = "port"
    ROUTE_TABLE_ID = "route-table-id"


class InfiniBand:
    CONFIG_SUBTREE = "infiniband"
    PKEY = "pkey"
    MODE = "mode"
    BASE_IFACE = "base-iface"
    DEFAULT_PKEY = 0xFFFF

    class Mode:
        DATAGRAM = "datagram"
        CONNECTED = "connected"


class MacVlan:
    TYPE = InterfaceType.MAC_VLAN
    CONFIG_SUBTREE = "mac-vlan"
    BASE_IFACE = "base-iface"
    MODE = "mode"
    PROMISCUOUS = "promiscuous"

    class Mode:
        UNKNOWN = "unknown"
        VEPA = "vepa"
        BRIDGE = "bridge"
        PRIVATE = "private"
        PASSTHRU = "passthru"
        SOURCE = "source"


class MacVtap(MacVlan):
    TYPE = InterfaceType.MAC_VTAP
    CONFIG_SUBTREE = "mac-vtap"


class Ieee8021X:
    CONFIG_SUBTREE = "802.1x"
    IDENTITY = "identity"
    EAP_METHODS = "eap-methods"
    PRIVATE_KEY = "private-key"
    PRIVATE_KEY_PASSWORD = "private-key-password"
    CLIENT_CERT = "client-cert"
    CA_CERT = "ca-cert"



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateValueError
from libnmstate.schema import LinuxBridge

from .bridge import BridgeIface
from .linux_bridge_port_vlan import NmstateLinuxBridgePortVlan

# The aging_time, forward_delay, hello_time, max_age options are multipled by
# 100(USER_HZ) when apply to kernel(via NM), so they are not impacted by this
# integer round up/down issue.
INTEGER_ROUNDED_OPTIONS = [
    LinuxBridge.Options.MULTICAST_LAST_MEMBER_INTERVAL,
    LinuxBridge.Options.MULTICAST_MEMBERSHIP_INTERVAL,
    LinuxBridge.Options.MULTICAST_QUERIER_INTERVAL,
    LinuxBridge.Options.MULTICAST_QUERY_RESPONSE_INTERVAL,
    LinuxBridge.Options.MULTICAST_STARTUP_QUERY_INTERVAL,
]

READ_ONLY_OPTIONS = [
    LinuxBridge.Options.HELLO_TIMER,
    LinuxBridge.Options.GC_TIMER,
]


class LinuxBridgeIface(BridgeIface):
    @property
    def _options(self):
        return self.raw.get(LinuxBridge.CONFIG_SUBTREE, {}).get(
            LinuxBridge.OPTIONS_SUBTREE, {}
        )

    def pre_edit_validation_and_cleanup(self):
        if self.is_up:
            self._validate()
            self._fix_vlan_filtering_mode()
        super().pre_edit_validation_and_cleanup()

    @property
    def port(self):
        return [p[LinuxBridge.Port.NAME] for p in self.port_configs]

    def _validate(self):
        self._validate_vlan_filtering_trunk_tags()
        self._validate_vlan_filtering_tag()
        self._validate_vlan_filtering_enable_native()

    def _validate_vlan_filtering_trunk_tags(self):
        for port_config in self.original_dict.get(
            LinuxBridge.CONFIG_SUBTREE, {}
        ).get(LinuxBridge.PORT_SUBTREE, []):
            port_vlan_state = port_config.get(
                LinuxBridge.Port.VLAN_SUBTREE, {}
            )
            vlan_mode = port_vlan_state.get(LinuxBridge.Port.Vlan.MODE)
            trunk_tags = port_vlan_state.get(
                LinuxBridge.Port.Vlan.TRUNK_TAGS, []
            )

            if vlan_mode == LinuxBridge.Port.Vlan.Mode.ACCESS:
                if trunk_tags:
                    raise NmstateValueError(
                        "Access port cannot have trunk tags"
                    )
            elif port_vlan_state:
                if not trunk_tags:
                    raise NmstateValueError(
                        "A trunk port needs to specify trunk tags"
                    )
            for trunk_tag in trunk_tags:
                _assert_vlan_filtering_trunk_tag(trunk_tag)

    def _validate_vlan_filtering_tag(self):
        """
        The "tag" is valid in access mode or tunk mode with
        "enable-native:True".
        """
        for port_config in self.original_dict.get(
            LinuxBridge.CONFIG_SUBTREE, {}
        ).get(LinuxBridge.PORT_SUBTREE, []):
            vlan_config = _get_port_vlan_config(port_config)
            if (
                vlan_config.get(LinuxBridge.Port.Vlan.TAG)
                and _vlan_is_trunk_mode(vlan_config)
                and not _vlan_is_enable_native(vlan_config)
            ):
                raise NmstateValueError(
                    "Tag cannot be use in trunk mode without enable-native"
                )

    def _validate_vlan_filtering_enable_native(self):
        for port_config in self.original_dict.get(
            LinuxBridge.CONFIG_SUBTREE, {}
        ).get(LinuxBridge.PORT_SUBTREE, []):
            vlan_config = _get_port_vlan_config(port_config)
            if _vlan_is_access_mode(vlan_config) and _vlan_is_enable_native(
                vlan_config
            ):
                raise NmstateValueError(
                    "enable-native cannot be set in access mode"
                )

    def _fix_vlan_filtering_mode(self):
        for port_config in self.port_configs:
            _vlan_config_clean_up(_get_port_vlan_config(port_config))

    def gen_metadata(self, ifaces):
        super().gen_metadata(ifaces)
        if not self.is_absent:
            for port_config in self.port_configs:
                ifaces.all_kernel_ifaces[
                    port_config[LinuxBridge.Port.NAME]
                ].update({BridgeIface.BRPORT_OPTIONS_METADATA: port_config})

    def remove_port(self, port_name):
        if self._bridge_config:
            self.raw[LinuxBridge.CONFIG_SUBTREE][LinuxBridge.PORT_SUBTREE] = [
                port_config
                for port_config in self.port_configs
                if port_config[LinuxBridge.Port.NAME] != port_name
            ]
        self.sort_port()

    def state_for_verify(self):
        self._normalize_bridge_port_vlan()
        self._remove_read_only_bridge_options()
        self._change_to_upper_group_addr()
        return super().state_for_verify()

    @staticmethod
    def is_integer_rounded(iface_state, current_iface_state):
        for key, value in iface_state._options.items():
            if key in INTEGER_ROUNDED_OPTIONS:
                try:
                    value = int(value)
                    cur_value = int(current_iface_state._options.get(key))
                except (TypeError, ValueError):
                    continue
                # With 250 HZ and 100 USER_HZ, every 8,000,000 will have 1
                # deviation, caused by:
                # * kernel set the value using clock_t_to_jiffies():
                #       jiffies = int(clock * 100 / 250)
                # * kernel showing the value using jiffies_to_clock_t():
                #       clock =  int(int(jiffies * ( (10 ** 9 + 250/2) / 250)
                #                    / 10 ** 9 * 100)
                #
                # The number 8,000,000 is found by exhaustion.
                # There is no good way to detect kernel HZ in user space. Hence
                # we check whether certain value is rounded.
                if cur_value != value:
                    if value >= 8 * (10 ** 6):
                        if abs(value - cur_value) <= int(
                            value / 8 * (10 ** 6)
                        ):
                            return key, value, cur_value
                    else:
                        if abs(value - cur_value) == 1:
                            return key, value, cur_value

        return None, None, None

    def _normalize_bridge_port_vlan(self):
        """
        * Set Bridge.Port.VLAN_SUBTREE as {} when not defined.
        * Expand VLAN ranges to single VLANs
        """
        for port_config in self.port_configs:
            if not port_config.get(LinuxBridge.Port.VLAN_SUBTREE):
                port_config[LinuxBridge.Port.VLAN_SUBTREE] = {}
            else:
                vlan_config = NmstateLinuxBridgePortVlan(
                    port_config[LinuxBridge.Port.VLAN_SUBTREE]
                ).to_dict(expand_vlan_range=True)
                port_config[LinuxBridge.Port.VLAN_SUBTREE] = vlan_config

    def _remove_read_only_bridge_options(self):
        for key in READ_ONLY_OPTIONS:
            self._bridge_config.get(LinuxBridge.OPTIONS_SUBTREE, {}).pop(
                key, None
            )

    def _change_to_upper_group_addr(self):
        cur_group_addr = self._bridge_config.get(
            LinuxBridge.OPTIONS_SUBTREE, {}
        ).get(LinuxBridge.Options.GROUP_ADDR)
        if cur_group_addr:
            self.raw[LinuxBridge.CONFIG_SUBTREE][LinuxBridge.OPTIONS_SUBTREE][
                LinuxBridge.Options.GROUP_ADDR
            ] = cur_group_addr.upper()


def _assert_vlan_filtering_trunk_tag(trunk_tag_state):
    vlan_id = trunk_tag_state.get(LinuxBridge.Port.Vlan.TrunkTags.ID)
    vlan_id_range = trunk_tag_state.get(
        LinuxBridge.Port.Vlan.TrunkTags.ID_RANGE
    )

    if vlan_id and vlan_id_range:
        raise NmstateValueError(
            "Trunk port cannot be configured by both id and range: {}".format(
                trunk_tag_state
            )
        )
    elif vlan_id_range:
        if not (
            {
                LinuxBridge.Port.Vlan.TrunkTags.MIN_RANGE,
                LinuxBridge.Port.Vlan.TrunkTags.MAX_RANGE,
            }
            <= set(vlan_id_range)
        ):
            raise NmstateValueError(
                "Trunk port range requires min / max keys: {}".format(
                    vlan_id_range
                )
            )


def _get_port_vlan_config(port_config):
    return port_config.get(LinuxBridge.Port.VLAN_SUBTREE, {})


# TODO: Group them into class _LinuxBridgePort
def _vlan_is_access_mode(vlan_config):
    return (
        vlan_config.get(LinuxBridge.Port.Vlan.MODE)
        == LinuxBridge.Port.Vlan.Mode.ACCESS
    )


def _vlan_is_trunk_mode(vlan_config):
    return (
        vlan_config.get(LinuxBridge.Port.Vlan.MODE)
        == LinuxBridge.Port.Vlan.Mode.TRUNK
    )


def _vlan_is_enable_native(vlan_config):
    return vlan_config.get(LinuxBridge.Port.Vlan.ENABLE_NATIVE) is True


def _vlan_config_clean_up(vlan_config):
    _vlan_remove_enable_native_if_access_mode(vlan_config)
    _vlan_remove_tag_if_trunk_mode_without_enable_native(vlan_config)
    _vlan_remove_trunk_tag_if_access_mode(vlan_config)


def _vlan_remove_enable_native_if_access_mode(vlan_config):
    if _vlan_is_access_mode(vlan_config):
        vlan_config.pop(LinuxBridge.Port.Vlan.ENABLE_NATIVE, None)


def _vlan_remove_tag_if_trunk_mode_without_enable_native(vlan_config):
    if _vlan_is_trunk_mode(vlan_config) and not _vlan_is_enable_native(
        vlan_config
    ):
        vlan_config.pop(LinuxBridge.Port.Vlan.TAG, None)


def _vlan_remove_trunk_tag_if_access_mode(vlan_config):
    if _vlan_is_access_mode(vlan_config):
        vlan_config.pop(LinuxBridge.Port.Vlan.TRUNK_TAGS, None)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.error import NmstateValueError
from libnmstate.schema import VRF

from .base_iface import BaseIface


class VrfIface(BaseIface):
    def sort_port(self):
        if self.port:
            self.raw[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE].sort()

    @property
    def route_table_id(self):
        return self.raw.get(VRF.CONFIG_SUBTREE, {}).get(VRF.ROUTE_TABLE_ID)

    @property
    def port(self):
        return self.raw.get(VRF.CONFIG_SUBTREE, {}).get(VRF.PORT_SUBTREE, [])

    @property
    def is_controller(self):
        return True

    @property
    def is_virtual(self):
        return True

    def pre_edit_validation_and_cleanup(self):
        super().pre_edit_validation_and_cleanup()
        if self.is_up:
            self._validate_route_table_id()

    def _validate_route_table_id(self):
        """
        The route table ID is manadatory
        """
        if not self.raw.get(VRF.CONFIG_SUBTREE, {}).get(VRF.ROUTE_TABLE_ID):
            raise NmstateValueError(
                f"Invalid route table ID for VRF interface {self.name}"
            )

    def remove_port(self, port_name):
        self.raw[VRF.CONFIG_SUBTREE][VRF.PORT_SUBTREE] = [
            s for s in self.port if s != port_name
        ]
        self.sort_port()



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from . import error
from . import schema

from .netapplier import apply
from .netapplier import commit
from .netapplier import rollback
from .netinfo import show
from .netinfo import show_running_config
from .nmstate import generate_configurations
from .prettystate import PrettyState
from .version import get_version as _get_version

__all__ = [
    "PrettyState",
    "apply",
    "commit",
    "error",
    "generate_configurations",
    "rollback",
    "schema",
    "show",
    "show_running_config",
]

__version__ = _get_version()



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from collections.abc import Mapping
from copy import deepcopy
import logging
from operator import itemgetter

from libnmstate.error import NmstateInternalError
from libnmstate.error import NmstateValueError
from libnmstate.iplib import is_ipv6_link_local_addr
from libnmstate.iplib import canonicalize_ip_address
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import InterfaceIPv6
from libnmstate.schema import InterfaceType
from libnmstate.schema import InterfaceState
from libnmstate.schema import LLDP
from libnmstate.schema import OvsDB
from libnmstate.schema import Ieee8021X

from ..state import state_match
from ..state import merge_dict


class IPState:
    def __init__(self, family, info):
        self._family = family
        self._info = info
        self._remove_stack_if_disabled()
        self._sort_addresses()
        self._canonicalize_ip_addr()
        self._canonicalize_dynamic()

    def _canonicalize_dynamic(self):
        if self.is_enabled and self.is_dynamic:
            self._info[InterfaceIP.ADDRESS] = []
            self._info.setdefault(InterfaceIP.AUTO_ROUTES, True)
            self._info.setdefault(InterfaceIP.AUTO_DNS, True)
            self._info.setdefault(InterfaceIP.AUTO_GATEWAY, True)
        else:
            for dhcp_option in (
                InterfaceIP.AUTO_ROUTES,
                InterfaceIP.AUTO_GATEWAY,
                InterfaceIP.AUTO_DNS,
            ):
                self._info.pop(dhcp_option, None)

    def _canonicalize_ip_addr(self):
        for addr in self.addresses:
            addr[InterfaceIP.ADDRESS_IP] = canonicalize_ip_address(
                addr[InterfaceIP.ADDRESS_IP]
            )

    def _sort_addresses(self):
        self.addresses.sort(key=itemgetter(InterfaceIP.ADDRESS_IP))

    def _remove_stack_if_disabled(self):
        if not self.is_enabled:
            self._info = {InterfaceIP.ENABLED: False}

    @property
    def is_enabled(self):
        return self._info.get(InterfaceIP.ENABLED, False)

    @property
    def is_dynamic(self):
        return self._info.get(InterfaceIP.DHCP) or self._info.get(
            InterfaceIPv6.AUTOCONF
        )

    @property
    def auto_dns(self):
        return self.is_dynamic and self._info.get(InterfaceIP.AUTO_DNS)

    @property
    def addresses(self):
        return self._info.get(InterfaceIP.ADDRESS, [])

    def to_dict(self):
        return deepcopy(self._info)

    def validate(self, original):
        if (
            self.is_enabled
            and self.is_dynamic
            and original.addresses
            and self.addresses
        ):
            logging.warning(
                f"Static addresses {original.addresses} "
                "are ignored when dynamic IP is enabled"
            )

    def remove_link_local_address(self):
        if self.addresses:
            self._info[InterfaceIP.ADDRESS] = [
                addr
                for addr in self.addresses
                if not is_ipv6_link_local_addr(
                    addr[InterfaceIP.ADDRESS_IP],
                    addr[InterfaceIP.ADDRESS_PREFIX_LENGTH],
                )
            ]


class BaseIface:
    CONTROLLER_METADATA = "_controller"
    CONTROLLER_TYPE_METADATA = "_controller_type"
    DNS_METADATA = "_dns"
    ROUTES_METADATA = "_routes"
    ROUTE_RULES_METADATA = "_route_rules"
    RULE_CHANGED_METADATA = "_changed"
    ROUTE_CHANGED_METADATA = "_changed"

    def __init__(self, info, save_to_disk=True):
        self._origin_info = deepcopy(info)
        self._info = deepcopy(info)
        self._is_desired = False
        self._is_changed = False
        self._name = self._info[Interface.NAME]
        self._save_to_disk = save_to_disk

    @property
    def can_have_ip_as_port(self):
        return False

    @property
    def is_user_space_only(self):
        """
        Whether this interface is user space only.
        User space network interface means:
            * Can have duplicate interface name against kernel network
              interfaces
            * Cannot be used subordinate of other kernel interfaces.
            * Due to limtation of nmstate, currently cannot be used as
              subordinate of other user space interfaces.
        """
        return False

    def sort_port(self):
        pass

    @property
    def raw(self):
        """
        Internal use only: Allowing arbitrary modifcation.
        """
        return self._info

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._info.get(Interface.TYPE, InterfaceType.UNKNOWN)

    @property
    def state(self):
        return self._info.get(Interface.STATE, InterfaceState.UP)

    @state.setter
    def state(self, value):
        self._info[Interface.STATE] = value

    @property
    def is_desired(self):
        return self._is_desired

    @property
    def is_changed(self):
        return self._is_changed

    def mark_as_changed(self):
        self._is_changed = True

    def mark_as_desired(self):
        self._is_desired = True

    def mark_as_absent_by_desire(self):
        self.mark_as_desired()
        self._info[Interface.STATE] = InterfaceState.ABSENT
        self._origin_info[Interface.STATE] = InterfaceState.ABSENT

    def to_dict(self):
        return deepcopy(self._info)

    @property
    def original_dict(self):
        return self._origin_info

    def ip_state(self, family):
        return IPState(family, self._info.get(family, {}))

    def is_ipv4_enabled(self):
        return self.ip_state(Interface.IPV4).is_enabled

    def is_ipv6_enabled(self):
        return self.ip_state(Interface.IPV6).is_enabled

    def is_dynamic(self, family):
        return self.ip_state(family).is_dynamic

    def pre_edit_validation_and_cleanup(self):
        """
        This function is called after metadata generation finished.
        Will do
         * Raise NmstateValueError when user desire(self.original_dict) is
           illegal.
         * Clean up illegal setting introduced by merging.
        We don't split validation from clean up as they might sharing the same
        check code.
        """
        if self.is_desired:
            if not self.is_absent:
                for family in (Interface.IPV4, Interface.IPV6):
                    self.ip_state(family).validate(
                        IPState(family, self._origin_info.get(family, {}))
                    )
                self._validate_port_ip()
                ip_state = self.ip_state(family)
                ip_state.remove_link_local_address()
                self._info[family] = ip_state.to_dict()
            if self.is_absent and not self._save_to_disk:
                self._info[Interface.STATE] = InterfaceState.DOWN

    def merge(self, other):
        self._ovsdb_pre_merge_clean_up(other)
        merge_dict(self._info, other._info)
        # If down state is not from orignal state, set it as UP.
        if (
            Interface.STATE not in self._origin_info
            and self.state == InterfaceState.DOWN
        ):
            self._info[Interface.STATE] = InterfaceState.UP

    def _ovsdb_pre_merge_clean_up(self, other):
        """
        * When user not define ovsdb external_ids, we copy from other.
        * When user define ovsdb external_ids, we don't merget from other and
          expecting user to provider full picture.
        """
        desired_external_ids = self._info.get(OvsDB.OVS_DB_SUBTREE, {}).get(
            OvsDB.EXTERNAL_IDS
        )
        current_external_ids = other._info.get(OvsDB.OVS_DB_SUBTREE, {}).get(
            OvsDB.EXTERNAL_IDS
        )
        if desired_external_ids is None:
            if current_external_ids:
                self._info.update(
                    {
                        OvsDB.OVS_DB_SUBTREE: {
                            OvsDB.EXTERNAL_IDS: current_external_ids
                        }
                    }
                )
        else:
            if current_external_ids:
                other._info[OvsDB.OVS_DB_SUBTREE].pop(OvsDB.EXTERNAL_IDS)

    def _validate_port_ip(self):
        for family in (Interface.IPV4, Interface.IPV6):
            ip_state = IPState(family, self._origin_info.get(family, {}))
            if (
                ip_state.is_enabled
                and self.controller
                and self.controller_type != InterfaceType.VRF
                and not self.can_have_ip_as_port
            ):
                raise NmstateValueError(
                    f"Interface {self.name} is port of {self.controller_type} "
                    f"interface {self.controller} which does not allow "
                    f"port to have {family} enabled"
                )

    @property
    def port(self):
        return []

    @property
    def parent(self):
        return None

    @property
    def need_parent(self):
        return False

    @property
    def is_absent(self):
        return self.state == InterfaceState.ABSENT

    @property
    def is_up(self):
        return self.state == InterfaceState.UP

    @property
    def is_ignore(self):
        return self.state == InterfaceState.IGNORE

    @property
    def is_down(self):
        return self.state == InterfaceState.DOWN

    def mark_as_up(self):
        self.raw[Interface.STATE] = InterfaceState.UP

    @property
    def is_controller(self):
        return False

    def set_controller(self, controller_iface_name, controller_type):
        self._info[BaseIface.CONTROLLER_METADATA] = controller_iface_name
        self._info[BaseIface.CONTROLLER_TYPE_METADATA] = controller_type
        if (
            not self.can_have_ip_as_port
            and controller_type != InterfaceType.VRF
        ):
            for family in (Interface.IPV4, Interface.IPV6):
                self._info[family] = {InterfaceIP.ENABLED: False}

    @property
    def controller(self):
        return self._info.get(BaseIface.CONTROLLER_METADATA)

    @property
    def controller_type(self):
        return self._info.get(BaseIface.CONTROLLER_TYPE_METADATA)

    def gen_metadata(self, ifaces):
        if self.is_controller and not self.is_absent:
            for port_name in self.port:
                port_iface = ifaces.all_kernel_ifaces[port_name]
                port_iface.set_controller(self.name, self.type)

    def update(self, info):
        self._info.update(info)

    @property
    def mac(self):
        return self._info.get(Interface.MAC)

    @property
    def mtu(self):
        return self._info.get(Interface.MTU)

    @mtu.setter
    def mtu(self, value):
        self._info[Interface.MTU] = value

    def _capitalize_mac(self):
        if self.mac:
            self._info[Interface.MAC] = self.mac.upper()

    def match(self, other):
        self_state = self.state_for_verify()
        other_state = other.state_for_verify()
        return state_match(self_state, other_state)

    def state_for_verify(self):
        """
        Return the network state as dictionary used for verifcation.
        Clean up if required.
        For BaseIface:
            * Capitalize MAC addresses.
            * Explicitly set state as UP if not defined.
            * Remove IPv6 link local addresses.
            * Remove empty description.
            * Change OVSDB value to string.
        """
        self._capitalize_mac()
        self.sort_port()
        for family in (Interface.IPV4, Interface.IPV6):
            ip_state = self.ip_state(family)
            ip_state.remove_link_local_address()
            self._info[family] = ip_state.to_dict()
        state = self.to_dict()
        _remove_empty_description(state)
        _remove_undesired_data(state, self.original_dict)
        _remove_lldp_neighbors(state)
        if Interface.STATE not in state:
            state[Interface.STATE] = InterfaceState.UP
        if self.is_absent and not self._save_to_disk:
            state[Interface.STATE] = InterfaceState.DOWN
        _convert_ovs_external_ids_values_to_string(state)

        return state

    def remove_port(self, port_name):
        if not self.is_controller:
            class_name = self.__class__.__name__
            raise NmstateInternalError(
                f"Invalid invoke of {class_name}.remove_port({port_name}) "
                f"as {class_name} is not a controller interface"
            )

    @property
    def is_virtual(self):
        return False

    def create_virtual_port(self, port_name):
        """
        When controller interface has non-exist port interface, controller
        should create virtual port for this name if possible, or else return
        None
        """
        return None

    def config_changed_port(self, _cur_iface):
        """
        Return a list of port interface name which has configuration changed
        compareing to cur_iface.
        """
        return []

    def store_dns_metadata(self, dns_metadata):
        for family, dns_config in dns_metadata.items():
            self.raw[family][BaseIface.DNS_METADATA] = dns_config

    def remove_dns_metadata(self):
        for family in (Interface.IPV4, Interface.IPV6):
            self.raw.get(family, {}).pop(BaseIface.DNS_METADATA, None)

    def store_route_metadata(self, route_metadata):
        for family, routes in route_metadata.items():
            try:
                self.raw[family][BaseIface.ROUTES_METADATA] = routes
            except KeyError:
                self.raw[family] = {BaseIface.ROUTES_METADATA: routes}

    def store_route_rule_metadata(self, route_rule_metadata):
        for family, rules in route_rule_metadata.items():
            self.raw[family][BaseIface.ROUTE_RULES_METADATA] = rules

    @property
    def copy_mac_from(self):
        return self._info.get(Interface.COPY_MAC_FROM)

    def apply_copy_mac_from(self, mac):
        """
        * Add MAC to original desire.
        * Remove Interface.COPY_MAC_FROM from original desire.
        * Update MAC of merge iface
        """
        self.raw[Interface.MAC] = mac
        self._origin_info[Interface.MAC] = mac
        self._origin_info.pop(Interface.COPY_MAC_FROM, None)

    @property
    def ieee_802_1x_conf(self):
        return self.raw.get(Ieee8021X.CONFIG_SUBTREE, {})


def _remove_empty_description(state):
    if state.get(Interface.DESCRIPTION) == "":
        del state[Interface.DESCRIPTION]


def _remove_lldp_neighbors(state):
    state.get(LLDP.CONFIG_SUBTREE, {}).pop(LLDP.NEIGHBORS_SUBTREE, None)


def _remove_undesired_data(state, desire):
    """
    For any key not defined in `desire`, remove it from `state`
    """
    key_to_remove = []
    for key, value in state.items():
        if key not in desire:
            key_to_remove.append(key)
        elif isinstance(value, Mapping):
            _remove_undesired_data(value, desire[key])
    for key in key_to_remove:
        state.pop(key)


def _convert_ovs_external_ids_values_to_string(iface_info):
    external_ids = iface_info.get(OvsDB.OVS_DB_SUBTREE, {}).get(
        OvsDB.EXTERNAL_IDS, {}
    )
    for key, value in external_ids.items():
        external_ids[key] = str(value)



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import logging
from operator import itemgetter

from libnmstate.ifaces import ovs
from libnmstate.ifaces.bridge import BridgeIface
from libnmstate.ifaces.ovs import OvsPortIface
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.schema import OVSBridge as OB
from libnmstate.schema import OVSInterface
from libnmstate.schema import OvsDB

from .common import NM


PORT_PROFILE_PREFIX = "ovs-port-"

CONTROLLER_TYPE_METADATA = "_controller_type"
CONTROLLER_METADATA = "_controller"
SETTING_OVS_EXTERNALIDS = "SettingOvsExternalIDs"
SETTING_OVS_EXTERNAL_IDS_SETTING_NAME = "ovs-external-ids"

NM_OVS_VLAN_MODE_MAP = {
    "trunk": OB.Port.Vlan.Mode.TRUNK,
    "access": OB.Port.Vlan.Mode.ACCESS,
    "native-tagged": OB.Port.Vlan.Mode.TRUNK,
    "native-untagged": OB.Port.Vlan.Mode.UNKNOWN,  # Not supported yet
    "dot1q-tunnel": OB.Port.Vlan.Mode.UNKNOWN,  # Not supported yet
}


class LacpValue:
    ACTIVE = "active"
    OFF = "off"


def has_ovs_capability(nm_client):
    return NM.Capability.OVS in nm_client.get_capabilities()


def create_bridge_setting(options_state):
    bridge_setting = NM.SettingOvsBridge.new()
    for option_name, option_value in options_state.items():
        if option_name == "fail-mode":
            if option_value:
                bridge_setting.props.fail_mode = option_value
        elif option_name == "mcast-snooping-enable":
            bridge_setting.props.mcast_snooping_enable = option_value
        elif option_name == "rstp":
            bridge_setting.props.rstp_enable = option_value
        elif option_name == "stp":
            bridge_setting.props.stp_enable = option_value

    return bridge_setting


def create_ovsdb_external_ids_setting(ovsdb_conf):
    if _is_nm_support_ovs_external_ids():
        nm_setting = getattr(NM, SETTING_OVS_EXTERNALIDS).new()
        for key, value in ovsdb_conf.get(OvsDB.EXTERNAL_IDS, {}).items():
            if not key.startswith("NM."):
                nm_setting.set_data(key, str(value))
        return nm_setting
    else:
        logging.warn(
            "Please upgrade NetworkManger to 1.30+ "
            "for the support OVS external ID modification"
        )
        return None


def create_port_setting(port_state):
    port_setting = NM.SettingOvsPort.new()

    lag_state = port_state.get(OB.Port.LINK_AGGREGATION_SUBTREE)
    if lag_state:
        mode = lag_state.get(OB.Port.LinkAggregation.MODE)
        if mode == OB.Port.LinkAggregation.Mode.LACP:
            port_setting.props.lacp = LacpValue.ACTIVE
        elif mode in (
            OB.Port.LinkAggregation.Mode.ACTIVE_BACKUP,
            OB.Port.LinkAggregation.Mode.BALANCE_SLB,
        ):
            port_setting.props.lacp = LacpValue.OFF
            port_setting.props.bond_mode = mode
        elif mode == OB.Port.LinkAggregation.Mode.BALANCE_TCP:
            port_setting.props.lacp = LacpValue.ACTIVE
            port_setting.props.bond_mode = mode

        down_delay = lag_state.get(OB.Port.LinkAggregation.Options.DOWN_DELAY)
        if down_delay:
            port_setting.props.bond_downdelay = down_delay
        up_delay = lag_state.get(OB.Port.LinkAggregation.Options.UP_DELAY)
        if up_delay:
            port_setting.props.bond_updelay = up_delay

    vlan_state = port_state.get(OB.Port.VLAN_SUBTREE, {})
    if OB.Port.Vlan.MODE in vlan_state:
        if vlan_state[OB.Port.Vlan.MODE] != OB.Port.Vlan.Mode.UNKNOWN:
            port_setting.props.vlan_mode = vlan_state[OB.Port.Vlan.MODE]
    if OB.Port.Vlan.TAG in vlan_state:
        port_setting.props.tag = vlan_state[OB.Port.Vlan.TAG]

    return port_setting


def create_interface_setting(patch_state):
    interface_setting = NM.SettingOvsInterface.new()
    settings = [interface_setting]

    if patch_state and patch_state.get(OVSInterface.Patch.PEER):
        interface_setting.props.type = "patch"
        settings.append(create_patch_setting(patch_state))
    else:
        interface_setting.props.type = "internal"

    return settings


def create_patch_setting(patch_state):
    patch_setting = NM.SettingOvsPatch.new()
    patch_setting.props.peer = patch_state[OVSInterface.Patch.PEER]

    return patch_setting


def get_ovs_bridge_info(nm_dev_ovs_br):
    iface_info = {OB.CONFIG_SUBTREE: {}}
    ports_info = _get_bridge_nmstate_ports_info(nm_dev_ovs_br)
    options = _get_bridge_options(nm_dev_ovs_br)

    if ports_info or options:
        iface_info[OB.CONFIG_SUBTREE] = {
            OB.PORT_SUBTREE: ports_info,
            OB.OPTIONS_SUBTREE: options,
        }
    return iface_info


def get_ovsdb_external_ids(nm_profile):
    iface_info = {}
    if _is_nm_support_ovs_external_ids():
        nm_setting = nm_profile.get_setting_by_name(
            SETTING_OVS_EXTERNAL_IDS_SETTING_NAME
        )
        if nm_setting:
            external_ids = {}
            for key in nm_setting.get_data_keys():
                external_ids[key] = nm_setting.get_data(key)
            iface_info[OvsDB.OVS_DB_SUBTREE] = {
                OvsDB.EXTERNAL_IDS: external_ids
            }
    return iface_info


def _get_bridge_nmstate_ports_info(nm_dev_ovs_br):
    ports_info = []
    for nm_dev_ovs_port in nm_dev_ovs_br.get_slaves():
        port_info = {}
        nm_dev_ovs_ifaces = nm_dev_ovs_port.get_slaves()
        if not nm_dev_ovs_ifaces:
            continue
        if len(nm_dev_ovs_ifaces) == 1:
            port_info[OB.Port.NAME] = nm_dev_ovs_ifaces[0].get_iface()
        else:
            port_info = _get_lag_nmstate_port_info(nm_dev_ovs_port)
        vlan_info = _get_vlan_info(nm_dev_ovs_port)
        if vlan_info:
            port_info[OB.Port.VLAN_SUBTREE] = vlan_info
        ports_info.append(port_info)
    return sorted(ports_info, key=itemgetter(OB.Port.NAME))


def get_interface_info(act_con):
    """
    Get OVS interface information from the NM profile.
    """
    info = {}
    if act_con:
        patch_setting = _get_patch_setting(act_con)
        if patch_setting:
            info[OVSInterface.PATCH_CONFIG_SUBTREE] = {
                OVSInterface.Patch.PEER: patch_setting.props.peer,
            }

    return info


def _get_patch_setting(act_con):
    """
    Get NM.SettingOvsPatch from NM.ActiveConnection.
    For any error, return None.
    """
    remote_con = act_con.get_connection()
    if remote_con:
        return remote_con.get_setting_ovs_patch()

    return None


def _get_lag_nmstate_port_info(nm_dev_ovs_port):
    OVS_LAG = OB.Port.LinkAggregation

    lag = {
        OVS_LAG.PORT_SUBTREE: sorted(
            [
                {OVS_LAG.Port.NAME: nm_dev_ovs_iface.get_iface()}
                for nm_dev_ovs_iface in nm_dev_ovs_port.get_slaves()
            ],
            key=itemgetter(OVS_LAG.Port.NAME),
        ),
    }
    mode = _get_lag_mode(nm_dev_ovs_port)
    if mode:
        lag[OVS_LAG.MODE] = mode
    return {
        OB.Port.NAME: nm_dev_ovs_port.get_iface(),
        OB.Port.LINK_AGGREGATION_SUBTREE: lag,
    }


def _get_lag_mode(nm_dev_ovs_port):
    """
    TODO: Use applied profile instead of on-disk one.
    """
    mode = None
    nm_setting = _get_nm_setting_ovs_port(nm_dev_ovs_port)
    if nm_setting:
        lacp = nm_setting.props.lacp
        mode = nm_setting.props.bond_mode
        if not mode:
            if lacp == LacpValue.ACTIVE:
                mode = OB.Port.LinkAggregation.Mode.LACP
            else:
                mode = OB.Port.LinkAggregation.Mode.ACTIVE_BACKUP
    return mode


def _get_vlan_info(nm_dev_ovs_port):
    nm_setting = _get_nm_setting_ovs_port(nm_dev_ovs_port)
    if nm_setting:
        vlan_mode = nm_setting.props.vlan_mode
        if vlan_mode:
            nmstate_vlan_mode = NM_OVS_VLAN_MODE_MAP.get(
                vlan_mode, OB.Port.Vlan.Mode.UNKNOWN
            )
            if nmstate_vlan_mode == OB.Port.Vlan.Mode.UNKNOWN:
                logging.warning(
                    f"OVS Port VLAN mode '{vlan_mode}' is not supported yet"
                )
                return {OB.Port.Vlan.MODE: OB.Port.Vlan.Mode.UNKNOWN}
            else:
                return {
                    OB.Port.Vlan.MODE: nmstate_vlan_mode,
                    OB.Port.Vlan.TAG: nm_setting.get_tag(),
                }
    return {}


def _get_nm_setting_ovs_port(nm_dev_ovs_port):
    nm_ac = nm_dev_ovs_port.get_active_connection()
    if nm_ac:
        nm_profile = nm_ac.props.connection
        if nm_profile:
            return nm_profile.get_setting(NM.SettingOvsPort)
    return None


def _get_bridge_options(bridge_device):
    bridge_options = {}
    bridge_profile = None
    act_conn = bridge_device.get_active_connection()
    if act_conn:
        bridge_profile = act_conn.props.connection

    if bridge_profile:
        bridge_setting = bridge_profile.get_setting(NM.SettingOvsBridge)
        bridge_options["stp"] = bridge_setting.props.stp_enable
        bridge_options["rstp"] = bridge_setting.props.rstp_enable
        bridge_options["fail-mode"] = bridge_setting.props.fail_mode or ""
        bridge_options[
            "mcast-snooping-enable"
        ] = bridge_setting.props.mcast_snooping_enable

    return bridge_options


def create_iface_for_nm_ovs_port(iface):
    iface_name = iface.name
    iface_info = iface.to_dict()
    port_options = iface_info.get(BridgeIface.BRPORT_OPTIONS_METADATA)
    if ovs.is_ovs_lag_port(port_options):
        port_name = port_options[OB.Port.NAME]
    else:
        port_name = PORT_PROFILE_PREFIX + iface_name
    return OvsPortIface(
        {
            Interface.NAME: port_name,
            Interface.TYPE: InterfaceType.OVS_PORT,
            Interface.STATE: iface.state,
            OB.OPTIONS_SUBTREE: port_options,
            CONTROLLER_METADATA: iface_info[CONTROLLER_METADATA],
            CONTROLLER_TYPE_METADATA: iface_info[CONTROLLER_TYPE_METADATA],
        }
    )


def _is_nm_support_ovs_external_ids():
    return hasattr(NM, SETTING_OVS_EXTERNALIDS)



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from operator import itemgetter

from libnmstate.schema import LinuxBridge as LB


class KernelBridgePortVlans:
    def __init__(
        self,
        vid_min=None,
        vid_max=None,
        is_pvid=False,
        is_egress_untagged=False,
    ):
        self.vid_min = vid_min
        self.vid_max = vid_max
        self.is_pvid = is_pvid
        self.is_egress_untagged = is_egress_untagged

    def get_vlan_tag_range(self):
        return (self.vid_min, self.vid_max)

    def to_dict(self):
        if self.vid_min != self.vid_max:
            return {
                LB.Port.Vlan.TrunkTags.ID_RANGE: {
                    LB.Port.Vlan.TrunkTags.MIN_RANGE: self.vid_min,
                    LB.Port.Vlan.TrunkTags.MAX_RANGE: self.vid_max,
                }
            }
        else:
            return {LB.Port.Vlan.TrunkTags.ID: self.vid_min}


class NmstateLinuxBridgePortVlan:
    def __init__(self, info):
        self._is_native = info.get(LB.Port.Vlan.ENABLE_NATIVE, False)
        self._trunk_tags = info.get(LB.Port.Vlan.TRUNK_TAGS, [])
        self._tag = info.get(LB.Port.Vlan.TAG, None)
        self._mode = info.get(LB.Port.Vlan.MODE, LB.Port.Vlan.Mode.ACCESS)

    @staticmethod
    def new_from_kernel_vlans(kernel_vlans):
        obj = NmstateLinuxBridgePortVlan({})
        obj._is_native = False
        trunk_tags = []

        is_access_port = NmstateLinuxBridgePortVlan._is_access_port(
            kernel_vlans
        )
        for kernel_vlan in kernel_vlans:
            vlan_min, vlan_max = kernel_vlan.get_vlan_tag_range()
            if vlan_min == 1 and vlan_max == 1:
                continue
            if is_access_port:
                obj._tag = vlan_min
            elif kernel_vlan.is_pvid and kernel_vlan.is_egress_untagged:
                obj._tag = vlan_max
                obj._is_native = True
            else:
                trunk_tags.append(kernel_vlan.to_dict())

        obj._trunk_tags = trunk_tags
        obj._mode = (
            LB.Port.Vlan.Mode.TRUNK if trunk_tags else LB.Port.Vlan.Mode.ACCESS
        )
        return obj

    def to_kernel_vlans(self):
        kernel_vlans = []
        if self._mode == LB.Port.Vlan.Mode.TRUNK:
            kernel_vlans.extend(
                [
                    NmstateLinuxBridgePortVlan._trunk_tag_to_kernel_vlan(
                        trunk_tag
                    )
                    for trunk_tag in self._trunk_tags
                ]
            )
            if self._is_native and self._tag:
                kernel_vlans.append(
                    KernelBridgePortVlans(
                        vid_min=self._tag,
                        vid_max=self._tag,
                        is_pvid=True,
                        is_egress_untagged=True,
                    )
                )
        elif self._mode == LB.Port.Vlan.Mode.ACCESS and self._tag:
            kernel_vlans.append(
                KernelBridgePortVlans(
                    vid_min=self._tag,
                    vid_max=self._tag,
                    is_pvid=True,
                    is_egress_untagged=True,
                )
            )
        return kernel_vlans

    @staticmethod
    def _trunk_tag_to_kernel_vlan(trunk_tag):
        vid_min = vid_max = trunk_tag.get(LB.Port.Vlan.TrunkTags.ID)
        if vid_min is None:
            ranged_vlan_tags = trunk_tag.get(LB.Port.Vlan.TrunkTags.ID_RANGE)
            vid_min = ranged_vlan_tags[LB.Port.Vlan.TrunkTags.MIN_RANGE]
            vid_max = ranged_vlan_tags[LB.Port.Vlan.TrunkTags.MAX_RANGE]

        return KernelBridgePortVlans(
            vid_min=vid_min,
            vid_max=vid_max,
            is_pvid=False,
            is_egress_untagged=False,
        )

    def to_dict(self, expand_vlan_range=False):
        trunk_tags = self._trunk_tags if self._trunk_tags else []

        if expand_vlan_range and trunk_tags:
            trunk_tags = _expand_trunk_tags(self._trunk_tags)

        port_vlan_info = {
            LB.Port.Vlan.MODE: self._mode,
            LB.Port.Vlan.TRUNK_TAGS: trunk_tags,
        }
        if self._tag:
            port_vlan_info[LB.Port.Vlan.TAG] = self._tag
        if self._mode == LB.Port.Vlan.Mode.TRUNK:
            port_vlan_info[LB.Port.Vlan.ENABLE_NATIVE] = self._is_native
        return port_vlan_info

    @staticmethod
    def _is_access_port(kernel_vlans):
        return (
            len(kernel_vlans) == 1
            and kernel_vlans[0].is_pvid
            and kernel_vlans[0].is_egress_untagged
        )


def _expand_trunk_tags(trunk_tags):
    expanded_trunk_tags = []
    for trunk_tag in trunk_tags:
        if LB.Port.Vlan.TrunkTags.ID_RANGE in trunk_tag:
            vid_range = trunk_tag[LB.Port.Vlan.TrunkTags.ID_RANGE]
            vid_min = vid_range[LB.Port.Vlan.TrunkTags.MIN_RANGE]
            vid_max = vid_range[LB.Port.Vlan.TrunkTags.MAX_RANGE]
            for vid in range(vid_min, vid_max + 1):
                expanded_trunk_tags.append({LB.Port.Vlan.TrunkTags.ID: vid})
        elif LB.Port.Vlan.TrunkTags.ID in trunk_tag:
            expanded_trunk_tags.append(
                {
                    LB.Port.Vlan.TrunkTags.ID: trunk_tag[
                        LB.Port.Vlan.TrunkTags.ID
                    ]
                }
            )
    expanded_trunk_tags.sort(key=itemgetter(LB.Port.Vlan.TrunkTags.ID))
    return expanded_trunk_tags



#
# Copyright (c) 2018-2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import copy
import logging

import jsonschema as js

from libnmstate.ifaces.ovs import is_ovs_running
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceType
from libnmstate.error import NmstateDependencyError

from . import schema
from .plugin import NmstatePlugin

MAX_SUPPORTED_INTERFACES = 1000


def schema_validate(data, validation_schema=schema.ifaces_schema):
    data = copy.deepcopy(data)
    _validate_max_supported_intface_count(data)
    for ifstate in data.get(schema.Interface.KEY, ()):
        if not ifstate.get(schema.Interface.TYPE):
            ifstate[schema.Interface.TYPE] = schema.InterfaceType.UNKNOWN
    js.validate(data, validation_schema)


def validate_capabilities(state, capabilities):
    validate_interface_capabilities(state.get(Interface.KEY, []), capabilities)


def validate_interface_capabilities(ifaces_state, capabilities):
    ifaces_types = {iface_state.get("type") for iface_state in ifaces_state}
    has_ovs_capability = NmstatePlugin.OVS_CAPABILITY in capabilities
    has_team_capability = NmstatePlugin.TEAM_CAPABILITY in capabilities
    ovs_is_running = is_ovs_running()
    for iface_type in ifaces_types:
        is_ovs_type = iface_type in (
            InterfaceType.OVS_BRIDGE,
            InterfaceType.OVS_INTERFACE,
            InterfaceType.OVS_PORT,
        )
        if is_ovs_type and not has_ovs_capability:
            if not ovs_is_running:
                raise NmstateDependencyError(
                    "openvswitch service is not started."
                )
            else:
                raise NmstateDependencyError(
                    "Open vSwitch NetworkManager support not installed "
                    "and started"
                )
        elif iface_type == InterfaceType.TEAM and not has_team_capability:
            raise NmstateDependencyError(
                "NetworkManager-team plugin not installed and started"
            )


def _validate_max_supported_intface_count(data):
    """
    Raises warning if the interfaces count in the single desired state
    exceeds the limit specified in MAX_SUPPORTED_INTERFACES
    """
    num_of_interfaces = len(
        [intface for intface in data.get(schema.Interface.KEY, ())]
    )
    if num_of_interfaces > MAX_SUPPORTED_INTERFACES:
        logging.warning(
            "Interfaces count exceeds the limit %s in desired state",
            MAX_SUPPORTED_INTERFACES,
        )



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from libnmstate.schema import RouteRule


NISPOR_RULE_ACTION_TABLE = "table"


def nispor_route_rule_state_to_nmstate(np_rules):
    return [
        _nispor_route_rule_to_nmstate(rl)
        for rl in np_rules
        if (rl.src or rl.dst) and rl.action == NISPOR_RULE_ACTION_TABLE
    ]


def _nispor_route_rule_to_nmstate(np_rl):
    rule = {
        RouteRule.ROUTE_TABLE: np_rl.table,
        RouteRule.PRIORITY: np_rl.priority
        if np_rl.priority
        else RouteRule.USE_DEFAULT_PRIORITY,
    }
    if np_rl.src:
        rule[RouteRule.IP_FROM] = np_rl.src
    if np_rl.dst:
        rule[RouteRule.IP_TO] = np_rl.dst
    return rule



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

from collections import defaultdict
import logging

from libnmstate.error import NmstateVerificationError
from libnmstate.error import NmstateValueError
from libnmstate.iplib import KERNEL_MAIN_ROUTE_TABLE_ID
from libnmstate.iplib import is_ipv6_address
from libnmstate.iplib import canonicalize_ip_network
from libnmstate.prettystate import format_desired_current_state_diff
from libnmstate.schema import Interface
from libnmstate.schema import InterfaceIP
from libnmstate.schema import RouteRule
from libnmstate.schema import Route

from .ifaces.base_iface import BaseIface
from .state import StateEntry
from .state import state_match


class RouteRuleEntry(StateEntry):
    def __init__(self, route_rule):
        self.ip_from = route_rule.get(RouteRule.IP_FROM)
        self.ip_to = route_rule.get(RouteRule.IP_TO)
        self.priority = route_rule.get(RouteRule.PRIORITY)
        self.route_table = route_rule.get(RouteRule.ROUTE_TABLE)
        self.state = route_rule.get(RouteRule.STATE)
        self._complement_defaults()
        self._canonicalize_ip_network()

    def _complement_defaults(self):
        if not self.absent:
            if self.ip_from is None:
                self.ip_from = ""
            if self.ip_to is None:
                self.ip_to = ""
            if self.priority is None:
                self.priority = RouteRule.USE_DEFAULT_PRIORITY
            if (
                self.route_table is None
                or self.route_table == RouteRule.USE_DEFAULT_ROUTE_TABLE
            ):
                self.route_table = KERNEL_MAIN_ROUTE_TABLE_ID

    def _canonicalize_ip_network(self):
        if self.ip_from:
            self.ip_from = canonicalize_ip_network(self.ip_from)
        if self.ip_to:
            self.ip_to = canonicalize_ip_network(self.ip_to)

    def _keys(self):
        return (self.ip_from, self.ip_to, self.priority, self.route_table)

    @property
    def is_ipv6(self):
        if self.ip_from:
            return is_ipv6_address(self.ip_from)
        elif self.ip_to:
            return is_ipv6_address(self.ip_to)
        else:
            logging.warning(
                f"Neither {RouteRule.IP_FROM} nor {RouteRule.IP_TO} "
                "is defined, treating it a IPv4 route rule"
            )
            return False

    @property
    def absent(self):
        return self.state == RouteRule.STATE_ABSENT

    def is_valid(self, config_iface_routes):
        """
        Return False when there is no route for defined route table.
        """
        found = False
        for route_set in config_iface_routes.values():
            for route in route_set:
                if route.table_id == self.route_table or (
                    route.table_id == Route.USE_DEFAULT_ROUTE_TABLE
                    and self.route_table == KERNEL_MAIN_ROUTE_TABLE_ID
                ):
                    found = True
                    break
        return found


class RouteRuleState:
    def __init__(self, route_state, des_rule_state, cur_rule_state):
        self._config_changed = False
        self._cur_rules = defaultdict(set)
        self._rules = defaultdict(set)
        if cur_rule_state:
            for entry in _get_config(cur_rule_state):
                rl = RouteRuleEntry(entry)
                self._cur_rules[rl.route_table].add(rl)
                if not route_state or rl.is_valid(
                    route_state.config_iface_routes
                ):
                    self._rules[rl.route_table].add(rl)
        if des_rule_state:
            self._merge_rules(des_rule_state, route_state)

    @property
    def _config(self):
        return _get_config(self._rules)

    def _merge_rules(self, des_rule_state, route_state):
        """
        Handle absent rules before adding desired rule entries to make sure
        absent rule does not delete rule defined in desired state.
        """
        for entry in _get_config(des_rule_state):
            rl = RouteRuleEntry(entry)
            if rl.absent:
                self._apply_absent_rules(rl)
        for entry in _get_config(des_rule_state):
            rl = RouteRuleEntry(entry)
            if not rl.absent:
                self._rules[rl.route_table].add(rl)

    def _apply_absent_rules(self, rl):
        """
        Remove rules based on absent rules and treat missing property as
        wildcard match.
        """
        absent_iface_table = rl.route_table
        for route_table, rule_set in self._rules.items():
            if absent_iface_table and absent_iface_table != route_table:
                continue
            new_rules = set()
            for rule in rule_set:
                if not rl.match(rule):
                    new_rules.add(rule)
            if new_rules != rule_set:
                self._rules[route_table] = new_rules

    def verify(self, cur_rule_state):
        current = RouteRuleState(
            route_state=None,
            des_rule_state=None,
            cur_rule_state=cur_rule_state,
        )
        for route_table, rules in self._rules.items():
            rule_info = [
                _remove_route_rule_default_values(r.to_dict())
                for r in sorted(rules)
            ]
            cur_rule_info = [
                r.to_dict()
                for r in sorted(current._rules.get(route_table, set()))
            ]

            if not state_match(rule_info, cur_rule_info):
                raise NmstateVerificationError(
                    format_desired_current_state_diff(
                        {RouteRule.KEY: {RouteRule.CONFIG: rule_info}},
                        {RouteRule.KEY: {RouteRule.CONFIG: cur_rule_info}},
                    )
                )

    @property
    def config_changed(self):
        return self._config_changed

    def gen_metadata(self, route_state, ifaces):
        """
        Generate metada which could used for storing into interface.
        Data structure returned is:
            {
                iface_name: {
                    Interface.IPV4: ipv4_route_rules,
                    Interface.IPV6: ipv6_route_rules,
                }
            }
        """
        route_rule_metadata = {}
        for route_table, rules in self._rules.items():
            iface_name = self._iface_for_route_table(
                route_state, route_table, ifaces
            )
            route_rule_metadata[iface_name] = {
                Interface.IPV4: [],
                Interface.IPV6: [],
            }
            if rules != self._cur_rules[route_table]:
                route_rule_metadata[iface_name][
                    BaseIface.RULE_CHANGED_METADATA
                ] = True
            for rule in rules:
                family = Interface.IPV6 if rule.is_ipv6 else Interface.IPV4
                route_rule_metadata[iface_name][family].append(rule.to_dict())
        return route_rule_metadata

    def _iface_for_route_table(self, route_state, route_table, ifaces):
        for routes in route_state.config_iface_routes.values():
            for route in routes:
                if route.table_id == route_table:
                    return route.next_hop_interface

        for iface in ifaces.values():
            autotable_ipv4 = (
                iface.to_dict()
                .get(Interface.IPV4, {})
                .get(InterfaceIP.AUTO_ROUTE_TABLE_ID)
            )
            autotable_ipv6 = (
                iface.to_dict()
                .get(Interface.IPV6, {})
                .get(InterfaceIP.AUTO_ROUTE_TABLE_ID)
            )
            if autotable_ipv4 == route_table or autotable_ipv6 == route_table:
                return iface.name

        raise NmstateValueError(
            "Failed to find interface to with route table ID "
            f"{route_table} to store route rules"
        )


def _get_config(state):
    return state.get(RouteRule.CONFIG, [])


def _remove_route_rule_default_values(rule):
    if rule.get(RouteRule.PRIORITY) == RouteRule.USE_DEFAULT_PRIORITY:
        del rule[RouteRule.PRIORITY]
    if rule.get(RouteRule.ROUTE_TABLE) == RouteRule.USE_DEFAULT_ROUTE_TABLE:
        del rule[RouteRule.ROUTE_TABLE]
    return rule



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

import copy

from libnmstate.error import NmstateVerificationError
from libnmstate.prettystate import format_desired_current_state_diff
from libnmstate.schema import DNS
from libnmstate.schema import Interface
from libnmstate.schema import Route
from libnmstate.schema import RouteRule

from .ifaces import Ifaces
from .dns import DnsState
from .route import RouteState
from .route_rule import RouteRuleState
from .state import state_match


class NetState:
    def __init__(
        self,
        desire_state,
        current_state=None,
        save_to_disk=True,
        gen_conf_mode=False,
    ):
        if current_state is None:
            current_state = {}
        self._ifaces = Ifaces(
            desire_state.get(Interface.KEY),
            current_state.get(Interface.KEY),
            save_to_disk,
            gen_conf_mode,
        )
        self._route = RouteState(
            self._ifaces,
            desire_state.get(Route.KEY),
            current_state.get(Route.KEY),
        )
        self._dns = DnsState(
            desire_state.get(DNS.KEY),
            current_state.get(DNS.KEY),
        )
        self._route_rule = RouteRuleState(
            self._route,
            desire_state.get(RouteRule.KEY),
            current_state.get(RouteRule.KEY),
        )
        self.desire_state = copy.deepcopy(desire_state)
        self.current_state = copy.deepcopy(current_state)
        if self.desire_state:
            self._ifaces.gen_dns_metadata(self._dns, self._route)
            self._ifaces.gen_route_metadata(self._route)
            self._ifaces.gen_route_rule_metadata(self._route_rule, self._route)

    def verify(self, current_state):
        self._ifaces.verify(current_state.get(Interface.KEY))
        self._dns.verify(current_state.get(DNS.KEY))
        self._route.verify(current_state.get(Route.KEY))
        self._route_rule.verify(current_state.get(RouteRule.KEY))
        self._verify_other_global_info(current_state)

    def _verify_other_global_info(self, current_state):
        for key, value in self.desire_state.items():
            if key not in (Interface.KEY, DNS.KEY, Route.KEY, RouteRule.KEY):
                cur_value = current_state.get(key)
                if not state_match(value, cur_value):
                    raise NmstateVerificationError(
                        format_desired_current_state_diff(
                            {key: value},
                            {key: cur_value},
                        )
                    )

    @property
    def ifaces(self):
        return self._ifaces

    @property
    def dns(self):
        return self._dns



#
# Copyright (c) 2020 Red Hat, Inc.
#
# This file is part of nmstate
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.


from libnmstate.ifaces import NmstateLinuxBridgePortVlan
from libnmstate.ifaces import KernelBridgePortVlans


def get_port_vlan_info(np_sub):
    return NmstateLinuxBridgePortVlan.new_from_kernel_vlans(
        [_np_vlan_to_kernel_vlan(np_vlan) for np_vlan in np_sub.vlans]
    ).to_dict()


def _np_vlan_to_kernel_vlan(np_vlan):
    if np_vlan.vid_range:
        vid_min, vid_max = np_vlan.vid_range
    else:
        vid_min = vid_max = np_vlan.vid
    return KernelBridgePortVlans(
        vid_min, vid_max, np_vlan.is_pvid, np_vlan.is_egress_untagged
    )
